#!/usr/bin/env python
# encoding: utf-8

"""Commandline parsing facility for Shredder.

Produces an option dict that can be used by shredder.application.
Some options are processed immediately however.
"""

# Stdlib:
import os
import sys
import logging

# External:
from gi.repository import Gio
from gi.repository import GLib


def show_version():
    """Print the version as shown by `rmlint --version`"""
    proc = Gio.Subprocess.new(
        ['rmlint', '--version'],
        Gio.SubprocessFlags.STDERR_PIPE
    )
    *_, data = proc.communicate_utf8()

    # Shredder version is always the same as rmlint.
    # So, let's just replace `rmlint` with `Shredder` :-)
    print(data.replace('rmlint', 'Shredder', 1), end='')
    sys.exit(-1)


def adjust_loglevel(root_logger, count):
    """Convert a -v count to a python loglevel."""
    count = max(0, min(4, count))

    root_logger.setLevel({
        0: logging.CRITICAL,
        1: logging.ERROR,
        2: logging.WARNING,
        3: logging.INFO,
        4: logging.DEBUG,
    }[count])


def parse_arguments(root_logger):
    """Parse the cmdline options using GOption."""
    sys.argv[0] = 'shredder'
    parser = GLib.option.OptionParser(
        "PATHS ...",
        description="A gui frontend to rmlint.",
        option_list=[
            GLib.option.make_option(
                "--add-location", "-a",
                type="filename",
                action="append",
                dest="locations",
                help="Add locations to locations view."
            ),
            GLib.option.make_option(
                "--scan", "-s",
                type="filename",
                action="append",
                dest="untagged",
                help="Add location to scan (as untagged path)."
            ),
            GLib.option.make_option(
                "--scan-tagged", "-S",
                type="filename",
                action="append",
                dest="tagged",
                help="Add location to scan (as tagged path)."
            ),
            GLib.option.make_option(
                "--load-script", "-l",
                type="filename",
                action="store",
                dest="script",
                help="Show `script` in editor view."
            ),
            GLib.option.make_option(
                "--verbose", "-v",
                action="count",
                dest='more_verbosity',
                help="Be more verbose."
            ),
            GLib.option.make_option(
                "--less-verbose", "-V",
                action="count",
                dest='less_verbosity',
                help="Be less verbose."
            ),
            GLib.option.make_option(
                "--show-settings", "-c",
                action="store_true",
                dest='show_settings',
                help="Show the settings view."
            ),
            GLib.option.make_option(
                "--version", "",
                action="store_true",
                dest="show_version",
                help="Show the version of Shredder."
            ),
        ]
    )

    try:
        parser.parse_args()
    except GLib.Error as err:
        root_logger.error(str(err))
        return None

    vals = parser.values
    if parser.values.show_version:
        show_version()

    adjust_loglevel(
        root_logger,
        (vals.more_verbosity or 0) +
        (vals.less_verbosity or 0) +
        4  # Default loglevel: info.
    )

    # Check paths to be valid:
    paths = (vals.tagged or []) + (vals.untagged or []) + [vals.script]
    for path in (path for path in paths if path):
        if not os.path.exists(path):
            root_logger.error('`%s` does not exist.', path)
            sys.exit(-1)

    return vals


if __name__ == '__main__':
    LOGGER = logging.getLogger('test-cmdline')
    print(parse_arguments(LOGGER))



#!/usr/bin/env python3
# encoding: utf-8
from nose import with_setup
from tests.utils import *

import subprocess

from parameterized import parameterized


def _check_interpreter(interpreter):
    try:
        subprocess.call([interpreter, "-c", "1 + 1"])
        return True
    except (subprocess.CalledProcessError, FileNotFoundError):
        return False


@parameterized(["python2", "python3"])
@with_setup(usual_setup_func, usual_teardown_func)
def test_paranoia(interpreter):
    if not _check_interpreter(interpreter):
        print(
            "Interpreter {} does not seem to be working, skipping test".format(
                interpreter
            )
        )
        return

    create_file('xxx', 'a')
    create_file('xxx', 'b')
    create_file('xxx', 'c')
    create_file('xxx', 'd')
    create_link('a', 'hardlink_a', symlink=False)

    head, *data, footer = run_rmlint(
        '-S a -o py:{t}/rmlint.py'.format(t=TESTDIR_NAME)
    )

    assert footer['duplicate_sets'] == 1
    assert footer['total_lint_size'] == 9
    assert footer['total_files'] == 5 # 1 is ignored as own output
    assert footer['duplicates'] == 4

    with open(os.path.join(TESTDIR_NAME, 'b'), 'w') as handle:
        handle.write('yyy')

    with open(os.path.join(TESTDIR_NAME, 'c'), 'w') as handle:
        handle.write('xxxx')

    text = subprocess.check_output([
        interpreter,
        os.path.join(TESTDIR_NAME, 'rmlint.py'),
        '-d',
        '-p'
    ])
    text = text.decode('utf-8')

    # subprocess.call('ls  -l ' + TESTDIR_NAME, shell=True)
    os.remove(os.path.join(TESTDIR_NAME, "rmlint.py"))
    head, *data, footer = run_rmlint(
        '-S a -o py:{t}/rmlint.py'.format(t=TESTDIR_NAME)
    )

    assert footer['duplicate_sets'] == 1
    assert footer['total_lint_size'] == 0
    assert footer['total_files'] == 4 # + 1
    assert footer['duplicates'] == 1

    assert 'Content differs' in text
    assert 'Size differs' in text
    assert 'Same inode' in text

    text = subprocess.check_output([
        interpreter,
        os.path.join(TESTDIR_NAME, 'rmlint.py'),
        '-d',
        '-p'
    ])
    os.remove(os.path.join(TESTDIR_NAME, "rmlint.py"))
    head, *data, footer = run_rmlint(
        '-S a -o py:{t}/rmlint.py'.format(t=TESTDIR_NAME)
    )

    # Nothing should change.
    assert footer['duplicate_sets'] == 1
    assert footer['total_lint_size'] == 0
    assert footer['total_files'] == 4 # + 1
    assert footer['duplicates'] == 1



#!/usr/bin/env python3
# encoding: utf-8
from nose import with_setup
from tests.utils import *


def filter_part_of_directory(data):
    return [entry for entry in data if entry["type"] != "part_of_directory"]

@with_setup(usual_setup_func, usual_teardown_func)
def test_simple():
    create_file('xxx', '1/a')
    create_file('xxx', '2/a')
    create_file('xxx', 'a')

    head, *data, footer = run_rmlint('-p -D --rank-by A')
    data = filter_part_of_directory(data)

    assert 2 == sum(find['type'] == 'duplicate_dir' for find in data)

    # One original, one dupe
    assert 1 == sum(find['type'] == 'duplicate_file' for find in data if find['is_original'])
    assert 1 == sum(find['type'] == 'duplicate_file' for find in data if not find['is_original'])
    assert data[0]['size'] == 3

    # -S A should sort in reverse lexigraphic order.
    assert data[0]['is_original']
    assert not data[1]['is_original']
    assert data[0]['path'].endswith('2')
    assert data[1]['path'].endswith('1')


@with_setup(usual_setup_func, usual_teardown_func)
def test_diff():
    create_file('xxx', '1/a')
    create_file('xxx', '2/a')
    create_file('xxx', '3/a')
    create_file('yyy', '3/b')
    head, *data, footer = run_rmlint('-p -D --rank-by A')
    data = filter_part_of_directory(data)

    assert 2 == sum(find['type'] == 'duplicate_dir' for find in data)
    assert data[0]['size'] == 3

    # -S A should sort in reverse lexigraphic order.
    assert data[0]['is_original']
    assert not data[1]['is_original']
    assert data[0]['path'].endswith('2')
    assert data[1]['path'].endswith('1')


@with_setup(usual_setup_func, usual_teardown_func)
def test_same_but_not_dupe():
    create_file('xxx', '1/a')
    create_file('xxx', '2/a')
    create_file('xxx', '2/b')
    head, *data, footer = run_rmlint('-p -D --rank-by A')
    data = filter_part_of_directory(data)

    # No duplicate dirs, but 3 duplicate files should be found.
    assert 0 == sum(find['type'] == 'duplicate_dir' for find in data)
    assert 3 == sum(find['type'] == 'duplicate_file' for find in data)

@with_setup(usual_setup_func, usual_teardown_func)
def test_hardlinks():
    create_file('xxx', '1/a')
    create_link('1/a', '1/link1')
    create_link('1/a', '1/link2')
    create_file('xxx', '2/a')
    create_link('2/a', '2/link1')
    create_link('2/a', '2/link2')

    head, *data, footer = run_rmlint('-p -D -l -S a')
    data = filter_part_of_directory(data)
    assert len(data) == 5
    assert data[0]['type'] == 'duplicate_dir'
    assert data[0]['path'].endswith('1')
    assert data[1]['type'] == 'duplicate_dir'
    assert data[1]['path'].endswith('2')

    # Hardlink duplicates:
    assert data[2]['type'] == 'duplicate_file'
    assert data[2]['path'].endswith('1/a')
    assert data[2]['is_original']
    assert data[3]['type'] == 'duplicate_file'
    assert data[3]['path'].endswith('1/link1')
    assert not data[3]['is_original']
    assert data[4]['type'] == 'duplicate_file'
    assert data[4]['path'].endswith('1/link2')
    assert not data[4]['is_original']

    head, *data, footer = run_rmlint('-D -S a -L')
    data = filter_part_of_directory(data)
    assert len(data) == 2
    assert data[0]['type'] == 'duplicate_file'
    assert data[0]['path'].endswith('a')
    assert data[1]['type'] == 'duplicate_file'
    assert data[1]['path'].endswith('a')


@with_setup(usual_setup_func, usual_teardown_func)
def test_deep_simple():
    create_file('xxx', 'deep/a/b/c/d/1')
    create_file('xxx', 'deep/e/f/g/h/1')
    head, *data, footer = run_rmlint('-D -S a')
    data = filter_part_of_directory(data)

    assert data[0]['path'].endswith('deep/a')
    assert data[1]['path'].endswith('deep/e')
    assert int(data[0]['checksum'], 16) > 0
    assert int(data[1]['checksum'], 16) > 0
    assert len(data) == 2


@with_setup(usual_setup_func, usual_teardown_func)
def test_deep_simple():
    create_file('xxx', 'd/a/1')
    create_file('xxx', 'd/b/empty')
    create_file('xxx', 'd/a/1')
    create_file('xxx', 'd/b/empty')
    head, *data, footer = run_rmlint('-p -D -S a')
    data = filter_part_of_directory(data)

    assert data[0]['path'].endswith('d/a')
    assert data[1]['path'].endswith('d/b')
    assert len(data) == 2


@with_setup(usual_setup_func, usual_teardown_func)
def test_dirs_with_empty_files_only():
    create_file('', 'a/empty')
    create_file('', 'b/empty')
    head, *data, footer = run_rmlint('-p -D -S a -T df,dd --size 0')
    data = filter_part_of_directory(data)

    assert len(data) == 2
    assert data[0]['path'].endswith('a')
    assert data[0]['type'] == "duplicate_dir"
    assert data[1]['path'].endswith('b')
    assert data[1]['type'] == "duplicate_dir"

    head, *data, footer = run_rmlint('-p -D -S a -T df,dd')
    data = filter_part_of_directory(data)
    assert len(data) == 0

    head, *data, footer = run_rmlint('-p -D -S a --size 0')
    data = filter_part_of_directory(data)
    assert len(data) == 2

    data.sort(key=lambda elem: elem["path"])
    assert data[0]['path'].endswith('a/empty')
    assert data[0]['type'] == "emptyfile"
    assert data[1]['path'].endswith('b/empty')
    assert data[1]['type'] == "emptyfile"


def create_nested(root, letters):
    summed = []
    for letter in letters:
        summed.append(letter)
        path = os.path.join(*([root] + summed + ['1']))
        create_file('xxx', path)


@with_setup(usual_setup_func, usual_teardown_func)
def test_deep_full():
    create_nested('deep', 'abcd')
    create_nested('deep', 'efgh')

    # subprocess.call('tree ' + TESTDIR_NAME, shell=True)
    # subprocess.call('./rmlint -p -S a -D ' + TESTDIR_NAME, shell=True)
    head, *data, footer = run_rmlint('-p -D -S a')
    data = filter_part_of_directory(data)

    assert len(data) == 6

    assert data[0]['path'].endswith('deep/a')
    assert data[0]['type'] == 'duplicate_dir'
    assert data[0]['is_original']
    assert data[1]['path'].endswith('deep/e')
    assert not data[1]['is_original']
    assert data[1]['type'] == 'duplicate_dir'

    for idx, ending in enumerate(['a/b/c/d/1', 'a/b/c/1', 'a/b/1', 'a/1']):
        assert data[idx + 2]['path'].endswith(ending)
        assert data[idx + 2]['type'] == 'duplicate_file'
        assert data[idx + 2]['is_original'] == (idx == 0)


@with_setup(usual_setup_func, usual_teardown_func)
def test_deep_full_twice():
    create_nested('deep_a', 'abcd')
    create_nested('deep_a', 'efgh')
    create_nested('deep_b', 'abcd')
    create_nested('deep_b', 'efgh')

    head, *data, footer = run_rmlint(
        '-D -S a {t}/deep_a {t}/deep_b'.format(
            t=TESTDIR_NAME
        ),
        use_default_dir=False
    )
    data = filter_part_of_directory(data)

    assert len(data) == 8

    assert data[0]['path'].endswith('deep_a')
    assert data[0]['type'] == 'duplicate_dir'
    assert data[0]['is_original']
    assert data[1]['path'].endswith('deep_b')
    assert data[1]['is_original'] == False
    assert data[1]['type'] == 'duplicate_dir'

    assert data[2]['path'].endswith('deep_a/a')
    assert data[2]['type'] == 'duplicate_dir'
    assert data[2]['is_original']
    assert data[3]['path'].endswith('deep_a/e')
    assert data[3]['is_original'] == False
    assert data[3]['type'] == 'duplicate_dir'

    for idx, ending in enumerate(['a/b/c/d/1', 'a/b/c/1', 'a/b/1', 'a/1']):
        assert data[idx + 4]['path'].endswith(ending)
        assert data[idx + 4]['type'] == 'duplicate_file'
        assert data[idx + 4]['is_original'] == (idx == 0)

    assert data[0]['path'].endswith('deep_a')
    assert data[0]['is_original']
    assert data[1]['path'].endswith('deep_b')
    assert not data[1]['is_original']
    assert data[2]['path'].endswith('deep_a/a')
    assert data[2]['is_original']
    assert data[3]['path'].endswith('deep_a/e')
    assert not data[3]['is_original']


@with_setup(usual_setup_func, usual_teardown_func)
def test_symlinks():
    create_file('xxx', 'a/z')
    create_link('a/z', 'a/x', symlink=True)
    create_file('xxx', 'b/z')
    create_link('b/z', 'b/x', symlink=True)

    head, *data, footer = run_rmlint('-p -D -S a -F')
    data = filter_part_of_directory(data)

    assert len(data) == 2
    assert data[0]['path'].endswith('z')
    assert data[0]['is_original']
    assert data[1]['path'].endswith('z')
    assert not data[1]['is_original']

    head, *data, footer = run_rmlint('-p -D -S a -f')
    data = filter_part_of_directory(data)

    assert len(data) == 2
    assert data[0]['path'].endswith('/a')
    assert data[0]['is_original']
    assert data[1]['path'].endswith('/b')
    assert not data[1]['is_original']


def mount_bind_teardown_func():
    if runs_as_root():
        subprocess.call(
            'umount {dst}'.format(
                dst=os.path.join(TESTDIR_NAME, 'a/b')
            ),
            shell=True
        )

    usual_teardown_func()


@with_setup(usual_setup_func, mount_bind_teardown_func)
def test_mount_binds():
    if not runs_as_root():
        return

    create_file('xxx', 'a/b/1')
    create_file('xxx', 'c/2')

    subprocess.call(
        'mount --rbind {src} {dst}'.format(
            src=TESTDIR_NAME,
            dst=os.path.join(TESTDIR_NAME, 'a/b')
        ),
        shell=True
    )
    create_file('xxx', 'a/3')

    head, *data, footer = run_rmlint('-S a')
    assert data[0]['path'].endswith('c/2')
    assert data[1]['path'].endswith('a/3')
    assert len(data) == 2


@with_setup(usual_setup_func, usual_teardown_func)
def test_keepall_tagged():
    # Test for Issue #141:
    # https://github.com/sahib/rmlint/issues/141
    #
    # Make sure -k protects duplicate directories too,
    # when they're in a pref'd path.
    create_file('test', 'origs/folder/subfolder/file')
    create_file('test', 'origs/samefolder/subfolder/file')
    create_file('test', 'dups/folder/subfolder/file')
    create_file('test', 'dups/samefolder/subfolder/file')
    create_file('abcd', 'unmatched/folder/subfolder/file')
    create_file('abcd', 'unmatched/samefolder/subfolder/unmatched')

    parentdir = TESTDIR_NAME
    dupedir = os.path.join(TESTDIR_NAME, 'dups')
    origdir = os.path.join(TESTDIR_NAME, 'origs')
    origsubdir = os.path.join(origdir, 'folder')
    unmatcheddir = os.path.join(TESTDIR_NAME, 'unmatched')

    def do_test(km_opts, untagged_path, tagged_path):
        options = '-D -S Ap {maybe_km} {untagged} // {tagged}'.format(
                maybe_km = km_opts,
                untagged = untagged_path,
                tagged = tagged_path)
        head, *data, footer = run_rmlint(options, use_default_dir=False)
        data = filter_part_of_directory(data)
        return [head, *data, footer]


    ### test 1: simple -km test
    head, *data, footer = do_test('-k -m', dupedir, origdir)

    assert len(data) >= 2
    assert footer['total_files'] == 4
    assert footer['duplicates'] == 2
    assert footer['duplicate_sets'] == 1

    assert data[0]['path'].endswith(origdir)
    assert data[0]['is_original']

    assert data[1]['path'].endswith(dupedir)
    assert not data[1]['is_original']

    ### test 2: -km test with tagged originals dir nested under untagged dir
    # Files in origdir are traversed as both untagged (as parentdir/origs) and
    # tagged (as origdir) but the tagged traversal should take precedence
    # during preprocessing path double removal.  Therefore should give same
    # result as previous, except for total file count.
    head, *data, footer = do_test('-k -m', parentdir, origdir)

    assert len(data) >= 2
    assert footer['duplicates'] == 2
    assert footer['duplicate_sets'] == 1

    assert data[0]['path'].endswith(origdir)
    assert data[0]['is_original']

    assert data[1]['path'].endswith(dupedir)
    assert not data[1]['is_original']

    ### test 3: tag just part of a nested originals dir
    head, *data, footer = do_test('-k -m', parentdir, origsubdir)
    assert len(data) == 4
    assert footer['duplicates'] == 3
    assert footer['duplicate_sets'] == 1

    ###  test 4: test that tagging takes precedence over -S Ap option
    head, *data, footer = do_test('', dupedir, origdir)
    assert len(data) == 4
    assert footer['total_files'] == 4
    assert footer['duplicates'] == 3
    assert footer['duplicate_sets'] == 1

    assert data[0]['path'].endswith(origdir)
    assert data[0]['is_original']

    assert data[1]['path'].endswith(dupedir)
    assert not data[1]['is_original']

    ### test 5: test self-duplicates in untagged dir are preserved by -m option
    head, *data, footer = do_test('-k -m', unmatcheddir, origdir)
    # unmatcheddir contains self-duplicates but is protected by -m
    # -o pretty (partial) output as at rmlint 82f433a:
    # ==> In total 4 files, whereof 0 are duplicates in 0 groups.

    assert len(data) == 0
    assert footer['total_files'] == 4
    assert footer['duplicates'] == 0
    assert footer['duplicate_sets'] == 0

    ### test 6: simple -KM test
    head, *data, footer = do_test('-K -M', origdir, dupedir)
    assert len(data) >= 2
    assert footer['total_files'] == 4
    assert footer['duplicates'] == 2
    assert footer['duplicate_sets'] == 1

    assert data[0]['path'].endswith(origdir)
    assert data[0]['is_original']

    assert data[1]['path'].endswith(dupedir)
    assert not data[1]['is_original']

    ### test 7: -KM test with tagged duplicates dir nested under untagged dir
    # Files in origdir are traversed as both untagged (as parentdir/origs) and
    # tagged (as origdir) but the tagged traversal should take precedence
    # during preprocessing path double removal.  Therefore should give same
    # result as previous, except for total file count.

    head, *data, footer = do_test('-K -M', parentdir, dupedir)
    assert len(data) >= 2
    assert footer['duplicates'] == 2
    assert footer['duplicate_sets'] == 1

    assert data[0]['path'].endswith(origdir)
    assert data[0]['is_original']

    assert data[1]['path'].endswith(dupedir)
    assert not data[1]['is_original']

    ### test 8: test self-duplicates in untagged dir are preserved by -m option
    # unmatcheddir contains self-duplicates but is protected by -M
    # -o pretty (partial) output as at rmlint 82f433a:
    # ==> In total 4 files, whereof 0 are duplicates in 0 groups.
    head, *data, footer = do_test('-K -M', origdir, unmatcheddir)

    assert len(data) == 0
    assert footer['total_files'] == 4
    assert footer['duplicates'] == 0
    assert footer['duplicate_sets'] == 0


@with_setup(usual_setup_func, usual_teardown_func)
def test_equal_content_different_layout():
    # Different duplicates in different subdirs.
    create_file('xxx', "tree-a/sub2/x")
    create_file('yyy', "tree-a/sub1/y")

    # Same files but on top level.
    create_file('xxx', "tree-b/x")
    create_file('yyy', "tree-b/y")

    head, *data, footer = run_rmlint('-p -D --rank-by a')
    data = filter_part_of_directory(data)

    assert data[0]["path"].endswith("tree-a")
    assert data[0]["is_original"] is True
    assert data[1]["path"].endswith("tree-b")
    assert data[1]["is_original"] is False

    # Now, try to honour the layout
    head, *data, footer = run_rmlint('-p -Dj --rank-by a')
    data = filter_part_of_directory(data)
    for point in data:
        assert point["type"] == "duplicate_file"


@with_setup(usual_setup_func, usual_teardown_func)
def test_nested_content_with_same_layout():
    create_nested('deep', 'xyzabc')
    create_nested('deep', 'uvwabc')

    head, *data, footer = run_rmlint('-Dj --rank-by a')
    data = filter_part_of_directory(data)

    assert len(data) == 10
    assert data[0]["path"].endswith("deep/u/v/w")
    assert data[1]["path"].endswith("deep/x/y/z")

    # No need to test again what the functions above already test,
    # just check if those are duplicate files as expected.
    for point in data[2:]:
        assert point["type"] == "duplicate_file"



#!/usr/bin/env python3
# encoding: utf-8
from nose import with_setup
from tests.utils import *


@with_setup(usual_setup_func, usual_teardown_func)
def test_cmdline():
    create_file('xxx', '1/a')
    # feed rmlint the same file twice via command line
    head, *data, footer = run_rmlint('{t}/1 {t}/1'.format(t=TESTDIR_NAME), use_default_dir=False)
    assert 0 == sum(find['type'] == 'duplicate_file' for find in data)

    head, *data, footer = run_rmlint('{t}/1/a {t}/1/a'.format(t=TESTDIR_NAME), use_default_dir=False)
    assert 0 == sum(find['type'] == 'duplicate_file' for find in data)

    head, *data, footer = run_rmlint('{t}/1 {t}/1/a'.format(t=TESTDIR_NAME), use_default_dir=False)
    assert 0 == sum(find['type'] == 'duplicate_file' for find in data)


@with_setup(usual_setup_func, usual_teardown_func)
def test_symlink_noloop():
    create_file('xxx', '1/a')
    create_link('1/a', '1/link', symlink=True)

    head, *data, footer = run_rmlint('{t}/1'.format(t=TESTDIR_NAME), use_default_dir=False)
    assert 0 == sum(find['type'] == 'duplicate_file' for find in data)

    head, *data, footer = run_rmlint('{t}/1 {t}/1/a'.format(t=TESTDIR_NAME), use_default_dir=False)
    assert 0 == sum(find['type'] == 'duplicate_file' for find in data)

    head, *data, footer = run_rmlint('{t}/1 {t}/1/link'.format(t=TESTDIR_NAME), use_default_dir=False)
    assert 0 == sum(find['type'] == 'duplicate_file' for find in data)

    head, *data, footer = run_rmlint('{t}/1/a {t}/1/link'.format(t=TESTDIR_NAME), use_default_dir=False)
    assert 0 == sum(find['type'] == 'duplicate_file' for find in data)

@with_setup(usual_setup_func, usual_teardown_func)
def test_symlink_loop():
    create_file('xxx', '1/a')
    create_link('1', '1/link', symlink=True)

    head, *data, footer = run_rmlint('{t}/1'.format(t=TESTDIR_NAME), use_default_dir=False)
    assert 0 == sum(find['type'] == 'duplicate_file' for find in data)

    head, *data, footer = run_rmlint('{t}/1 {t}/1/link'.format(t=TESTDIR_NAME), use_default_dir=False)
    assert 0 == sum(find['type'] == 'duplicate_file' for find in data)


def mount_bind_teardown_func():
    if runs_as_root():
        subprocess.call(
            'umount {dst}'.format(
                dst=os.path.join(TESTDIR_NAME, 'a/b')
            ),
            shell=True
        )

    usual_teardown_func()

@with_setup(usual_setup_func, mount_bind_teardown_func)
def test_mount_binds():
    if not runs_as_root():
        return

    create_file('xxx', 'a/b/1')
    create_file('xxx', 'c/2')

    subprocess.call(
        'mount --rbind {src} {dst}'.format(
            src=TESTDIR_NAME,
            dst=os.path.join(TESTDIR_NAME, 'a/b')
        ),
        shell=True
    )
    create_file('xxx', 'a/3')

    head, *data, footer = run_rmlint('{t} {t}/a/b -S pa'.format(t=TESTDIR_NAME), use_default_dir=False)

    assert 3 == sum(find['type'] == 'duplicate_file' for find in data)

    # the actual order is a bit difficult to pin down since files 2
    # and 3 can be reached 2 different ways:
    # /tmp/rmlint-unit-testdir
    # ├── a
    # │   ├── 3*
    # │   └── b
    # │       ├── a
    # │       │   ├── 3
    # │       │   └── b
    # │       │       └── 1
    # │       └── c
    # │           └── 2
    # └── c
    #     └── 2*
    assert data[0]['path'].endswith('/2')
    assert data[1]['path'].endswith('/3')
    assert len(data) == 3




#!/usr/bin/env python3
# encoding: utf-8

#############
# UTILITIES #
#############

import os
import json
import time
import pprint
import shutil
import shlex
import struct
import subprocess
import contextlib
import xattr


TESTDIR_NAME = os.getenv('RM_TS_DIR') or '/tmp/rmlint-unit-testdir'

CKSUM_TYPES = [
    'murmur',
    'metro',
    'metro256',
    'md5',
    'sha1',
    'sha256',
    'sha3-256',
    'sha3-384',
    'sha3-512',
    'blake2s',
    'blake2b',
    'blake2sp',
    'blake2bp',
    'xxhash',
    'highway64',
    'highway128',
    'highway256',
    #'cumulative',
    #'ext',
    'paranoid',
]

def runs_as_root():
    return os.geteuid() == 0


def get_env_flag(name):
    try:
        return int(os.environ.get(name) or 0)
    except ValueError:
        print('{n} should be an integer.'.format(n=name))


def use_valgrind():
    return get_env_flag('RM_TS_USE_VALGRIND')


def keep_testdir():
    return get_env_flag('RM_TS_KEEP_TESTDIR')


def create_testdir(*extra_path):
    try:
        path = TESTDIR_NAME
        if extra_path:
            path = os.path.join(TESTDIR_NAME, *extra_path)
        os.makedirs(path)
    except OSError:
        pass


def which(program):
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath and is_exe(program):
        return program
    else:
        for path in (os.environ.get("PATH") or []).split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None


def has_feature(feature):
    return ('+' + feature) in subprocess.check_output(
        ['./rmlint', '--version'], stderr=subprocess.STDOUT
    ).decode('utf-8')


RMLINT_BINARY_DIR = os.getcwd()


def run_rmlint_once(*args,
                    dir_suffix=None,
                    use_default_dir=True,
                    outputs=None,
                    with_json=True,
                    directly_return_output=False,
                    use_shell=False,
                    verbosity="-V"):
    if use_default_dir:
        if dir_suffix:
            target_dir = os.path.join(TESTDIR_NAME, dir_suffix)
        else:
            target_dir = TESTDIR_NAME
    else:
        target_dir = ""

    if use_valgrind():
        env = {
            'G_DEBUG': 'gc-friendly',
            'G_SLICE': 'always-malloc'
        }
        cmd = [which('valgrind'), '--error-exitcode=1', '-q']
        if get_env_flag('RM_TS_CHECK_LEAKS'):
            cmd.extend( ['--leak-check=full', '--show-leak-kinds=definite', '--errors-for-leak-kinds=definite'] )
    elif get_env_flag('RM_TS_USE_GDB'):
        env, cmd = {}, ['/usr/bin/gdb', '-batch', '--silent', '-ex=run', '-ex=thread apply all bt', '-ex=quit', '--args']
    else:
        env, cmd = {}, []

    cmd = [os.path.join(RMLINT_BINARY_DIR, "rmlint"), verbosity]
    if target_dir:
        cmd.append(target_dir)

    cmd += shlex.split(' '.join(args))
    if with_json:
        cmd += ['-o', 'json:/tmp/out.json', '-c', 'json:oneline']

    for idx, output in enumerate(outputs or []):
        cmd.append('-o')
        cmd.append('{f}:{p}'.format(
            f=output, p=os.path.join(TESTDIR_NAME, '.' + output + '-' + str(idx)))
        )

    # filter empty strings
    cmd = list(filter(None, cmd))

    if get_env_flag('RM_TS_PRINT_CMD'):
        print('Running cmd from `{cwd}`: {cmd}'.format(
            cwd=os.getcwd(),
            cmd=' '.join(cmd)
        ))

    if get_env_flag('RM_TS_SLEEP'):
        print('Waiting for 1000 seconds.')
        time.sleep(1000)

    if use_shell is True:
        # Use /bin/bash, not /bin/sh
        cmd = ["/bin/bash", "-c", " ".join(cmd)]

    output = subprocess.check_output(cmd, env=env)
    if get_env_flag('RM_TS_USE_GDB'):
        print('==> START OF GDB OUTPUT <==')
        print(output.decode('utf-8'))
        print('==> END OF GDB OUTPUT <==')

    if directly_return_output:
        return output

    if with_json:
        with open('/tmp/out.json', 'r') as f:
            json_data = json.loads(f.read())
    else:
        json_data = []

    read_outputs = []
    for idx, output in enumerate(outputs or []):
        with open(os.path.join(TESTDIR_NAME, '.' + output + '-' + str(idx)), 'r', encoding='utf8') as handle:
            read_outputs.append(handle.read())
    if outputs is None:
        return json_data
    else:
        return json_data + read_outputs


def compare_json_doc(doc_a, doc_b, compare_checksum=False):
    keys = [
        'disk_id', 'inode', 'mtime', 'path', 'size', 'type'
    ]

    if compare_checksum and 'checkum' in doc_a and 'checksum' in doc_b:
        keys.append('checksum')

    for key in keys:
        # It's okay for unfinished checksums to have some missing fields.
        if doc_a['type'] == doc_b['type'] == 'unique_file':
            continue

        if doc_a[key] != doc_b[key]:
            print('  !! Key differs: ', key, doc_a[key], '!=', doc_b[key])
            return False

    return True


def compare_json_docs(docs_a, docs_b, compare_checksum=False):
    paths_a, paths_b = {}, {}

    for doc_a in docs_a[1:-1]:
        paths_a[doc_a['path']] = doc_a

    for doc_b in docs_b[1:-1]:
        paths_b[doc_b['path']] = doc_b

    for path_a, doc_a in paths_a.items():
        # if path_a not in paths_b:
        #     print('####', doc_a, path_a, '\n', docs_b, '\n\n', list(paths_b))
        doc_b = paths_b[path_a]
        if not compare_json_doc(doc_a, doc_b, compare_checksum):
            print('!! OLD:')
            pprint.pprint(doc_a)
            print('!! NEW:')
            pprint.pprint(doc_b)
            print('------- DIFF --------')
            return False

    if docs_a[-1] != docs_b[-1]:
        print('!! FOOTER DIFFERS', docs_a[-1], docs_b[-1])
        return False

    return True


def run_rmlint_pedantic(*args, **kwargs):
    options = [
        '--with-fiemap',
        '--without-fiemap',
        '--fake-pathindex-as-disk',
        '--fake-fiemap',
        '-P',
        '-PP',
        '--limit-mem 1M --algorithm=paranoid',
        '--buffered-read',
        '--threads=1',
        '--shred-never-wait',
        '--shred-always-wait',
        '--no-mount-table'
    ]


    # Note: sha512 is supported on all system which have
    #       no recent enough glib with. God forsaken debian people.
    if has_feature('sha512'):
        CKSUM_TYPES.append('sha512')

    if has_feature('sse4'):
        CKSUM_TYPES.append('metrocrc')
        CKSUM_TYPES.append('metrocrc256')

    for cksum_type in CKSUM_TYPES:
        options.append('--algorithm=' + cksum_type)

    data = None

    output_len = len(kwargs.get('outputs', []))

    for option in options:
        new_data = run_rmlint_once(*(args + (option, )), **kwargs)

        data_skip, new_data_skip = data, new_data

        if output_len != 0:
            if new_data:
                new_data_skip = new_data[:-output_len]

            if data:
                data_skip = data[:-output_len]

        # We cannot compare checksum in all cases.
        compare_checksum = not option.startswith('--algorithm=')

        if data_skip is not None and not 'directly_return_output' in kwargs and not compare_json_docs(data_skip, new_data_skip, compare_checksum):
            pprint.pprint(data_skip)
            pprint.pprint(new_data_skip)
            raise AssertionError("Optimisation too optimized: " + option)

        data = new_data

    return data


def run_rmlint(*args, force_no_pendantic=False, **kwargs):
    if get_env_flag('RM_TS_PEDANTIC') and force_no_pendantic is False:
        return run_rmlint_pedantic(*args, **kwargs)
    else:
        return run_rmlint_once(*args, **kwargs)


def create_dirs(path):
    full_path = os.path.join(TESTDIR_NAME, path)

    try:
        os.makedirs(full_path)
    except OSError:
        pass

    return full_path


def create_link(path, target, symlink=False):
    f = os.symlink if symlink else os.link
    f(
        os.path.join(TESTDIR_NAME, path),
        os.path.join(TESTDIR_NAME, target)
    )


def create_file(data, name, mtime=None, write_binary=False):
    full_path = os.path.join(TESTDIR_NAME, name)
    if '/' in name:
        try:
            os.makedirs(os.path.dirname(full_path))
        except OSError:
            pass

    with open(full_path, 'wb' if write_binary else 'w') as handle:
        if write_binary:
            if isinstance(data, int):
                handle.write(struct.pack('i', data))
            else:
                assert False, "Unhandled data type for binary write: " + data
        else:
            handle.write(data)

    if not mtime is None:
        subprocess.call(['touch', '-m', '-d', str(mtime), full_path])

    return full_path


def warp_file_to_future(name, seconds):
    now = time.time()
    os.utime(os.path.join(TESTDIR_NAME, name), (now + 2, now + 2))


def usual_setup_func():
    shutil.rmtree(path=TESTDIR_NAME, ignore_errors=True)
    create_testdir()


def usual_teardown_func():
    if not keep_testdir():
        # Allow teardown to be called more than once:
        try:
            shutil.rmtree(path=TESTDIR_NAME)
        except OSError:
            pass


@contextlib.contextmanager
def create_special_fs(name, fs_type='ext4'):
    """
    Used to create a special filesystem container in TESTDIR_NAME
    under »name«. The type of the filesystem will be »fs_type« (as long
    we have a »mkfs.xyz« binary for that.) This method needs root privileges.

    Returns: The path of the created directory.
    """
    mount_path = os.path.join(TESTDIR_NAME, name)
    device_path = mount_path + ".device"

    commands = [
        "dd if=/dev/zero of={} bs=1M count=20".format(device_path),
        "mkdir -p {}".format(mount_path),
        "mkfs.{} {}".format(fs_type, device_path),
        "mount -o loop {} {}".format(device_path, mount_path),
    ]

    for command in commands:
        subprocess.run(
                command,
                shell=True,
                check=True,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
        )

    try:
        yield mount_path
    finally:
        # whatever happens: unmount it again.
        # we'll get test errors in the next tests otherwise.
        unmount_command = "umount {}".format(mount_path)
        subprocess.run(unmount_command, shell=True, check=True)


def must_read_xattr(path):
    """
    Read all extended attributes of a »path«.

    NOTE: This will only work on non-tmpfs mounts.
          See create_special_fs for a workaround.
    """
    return dict(xattr.xattr(os.path.join(TESTDIR_NAME, path)).items())



#!/usr/bin/env python
# encoding: utf-8

"""
GtkTreeModel Implementation for a directory tree and accompanying classes.

We use an own Model implementation instead of GtkTreeStore for sake of clean
code and more fine-grained control over the performance.

Since this is pure python the performance might not be 100% optimal but we try
to make the ui rendering as fluid as possible by making most of the costly
operations non-blocking.

Most of the code should be very clean but some vodoo is hidden.
If you don't find it, that's good. If you do, blame @sahib.
"""

# Stdlib:
import os
import time
import logging

from collections import OrderedDict, defaultdict, deque

# External:
from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject

# Internal:
from shredder.util import CellRendererSize
from shredder.util import CellRendererModifiedTime
from shredder.util import CellRendererCount
from shredder.util import CellRendererLint
from shredder.util import PopupMenu, NodeState

from shredder.query import Query


LOGGER = logging.getLogger('tree')


# Unique ID used for PathTreeModel's GtkTreeIters:
# Useful for debugging, sometimes iters will
# not come from our own model and may be invalid.
PATH_MODEL_STAMP = 0xDEAD

# Process at max work items before letting the
# mainloop run for a short time. This is to prevent
# lagging in the user interface.
# This does not speed up the operation itself of course.
PATH_MODEL_CHUNK_SIZE = 500

# Wait this much ms before processing the next chunk.
PATH_MODEL_TIMEOUT_MS = 50


class Column:
    """Column Enumeration to avoid using direct indices.

    Only this class needs to be changed when adding/modifying columns.
    """
    SIZE, COUNT, MTIME, TAG, CKSUM, PATH, TOOLTIP = range(7)
    TYPES = [float, int, int, int, int, str, str]

    @staticmethod
    def make_row(md_map):
        """Convert an rmlint json dict to a tree row"""
        is_original = md_map.get('is_original', False)
        if md_map.get('type', '').startswith('duplicate_'):
            if is_original:
                tag = NodeState.ORIGINAL
            else:
                tag = NodeState.DUPLICATE
        else:
            tag = NodeState.NONE

        # Use a list so we can update the counts and size later:
        # Note1: PATH and TOOLTIP are not included.
        #        Those are generated on demand.
        # Note2: Twins is negative, to differentiate between
        #        directory count and number of twins (!= 0)
        return [
            md_map.get('size', 0),
            -md_map.get('twins', 0),
            md_map.get('mtime', 0),
            tag,
            md_map.get('checksum', '')
        ]


class PathNode:
    """A single node in the PathTrie."""
    # Save a bit of memory:
    __slots__ = [
        # Public:
        'name', 'parent', 'children', 'row',

        # Private:
        'is_leaf', 'idx', 'indices', 'depth'
    ]

    def __init__(self, name, parent, metadata=None, children=None):
        # Public:
        self.name = name
        self.parent = parent
        self.children = children or OrderedDict()

        # If None probably a directory node, show the bare minimum:
        self.row = Column.make_row(metadata or {})

        # Private:
        self.indices = deque()
        self.idx = 0
        self.is_leaf = False
        self.depth = (parent.depth + 1) if parent else 0

    def __getitem__(self, idx):
        """Get a column value by its column index.

        Some values might get freshly constructed on this call.
        """
        if idx is Column.PATH:
            return self.name
        elif idx is Column.TOOLTIP:
            return GLib.markup_escape_text(self.build_path(), -1)
        else:
            return self.row[idx]

    def append(self, name):
        """Append a node as child of this node.

        If is_leaf is True the size/count of all intermediate
        non-leaf nodes above are updated.
        """
        node = PathNode(name, self)
        node.idx = len(self.indices)

        self.children[name] = node
        self.indices.append(node)

        return node

    def make_leaf(self, row):
        """Convert the node to a leaf node.

        Metadata will be copied to the node and intermediate
        directories above will be updated with count & size.
        """
        self.is_leaf = True
        self.row = row

        # Update intermediate directories top of it:
        for parent in (node for node in self.up() if not node.is_leaf):
            parent.row[Column.COUNT] += 1
            parent.row[Column.SIZE] += row[Column.SIZE]

    def up(self):
        """Iterate the trie up to root."""
        yield self
        if self.parent is not None:
            yield from self.parent.up()

    def build_path(self):
        """Recursively build the absolute path of this node"""
        return os.path.join(*reversed([n.name for n in self.up()]))

    def build_iter_path(self):
        """Recursively build an iter path suitable for GtkTreePath"""
        return list(reversed([n.idx for n in self.up()]))

    def neighbor(self, offset):
        """Get the neighbor of this node by a certain offset"""
        if self.parent is None:
            return None

        neighbor_idx = self.idx + offset
        if 0 <= neighbor_idx < len(self.parent.indices):
            return self.parent.indices[neighbor_idx]

        return None


def _create_root_path_index(index, path, node):
    """Create a (trie-like) recursive dict as fast root path lookup."""
    curr_map, last_map, name = index, None, ''

    for name in (comp for comp in path.split('/') if comp):
        last_map = curr_map
        curr_map = curr_map.setdefault(name, {})

    last_map[name] = node


def _lookup_root_path_index(index, components):
    """Lookup a node in the recursive dict by a split path,
    if found the node is returned and components is modified.
    """
    curr_map = index
    for idx, name in enumerate(components):
        curr_map = curr_map.get(name)
        if curr_map is None:
            return None

        if isinstance(curr_map, PathNode):
            del components[:idx + 1]
            return curr_map

    return None


class PathTrie(GObject.Object):
    """Python version of rmlint's pathtricia trie."""
    __gsignals__ = {
        'node-updated': (GObject.SIGNAL_RUN_FIRST, None, (GObject.TYPE_UINT64, ))
    }

    def __init__(self, root_paths=None):
        GObject.Object.__init__(self)

        self.root = PathNode('/', None, {})
        self.sub_roots = []
        self.nodes = {id(self.root): self.root}
        self.max_depth = 0

        self.root_paths = {}
        for root_path in root_paths or []:
            # Append the sub root node manually:
            sub_root_node = self.root.append(root_path.strip('/'))
            self.nodes[id(sub_root_node)] = sub_root_node
            self.sub_roots.append(sub_root_node)

            # Also add it to the "special" index.
            _create_root_path_index(self.root_paths, root_path, sub_root_node)

        self._groups = defaultdict(list)

    def __iter__(self):
        return self.iterate(None)

    def __len__(self):
        return len(self.nodes)

    def __repr__(self):
        """Return a simple string version of the trie"""
        view = []
        for node in self:
            view.append((' ' * node.depth * 2) + node.name)

        return '\n'.join(view)

    def __getitem__(self, path):
        return self.find(path)

    def __setitem__(self, path, value):
        self.insert(path, value)

    def iterate(self, node=None):
        """Iterate trie down from node.
        If node is None, root is assumed;
        """
        node = node or self.root
        yield node

        for child in node.indices:
            yield from self.iterate(child)

    def lookup_node_id(self, node_id):
        """Lookup a node by its id."""
        return self.nodes.get(node_id)

    def update_node(self, node, column_id, value):
        """Update a PathNode *and* emit a node-updated signal."""
        node.row[column_id] = value
        self.emit('node-updated', id(node))

    def group(self, cksum):
        """Get a list of nodes that have the same checksum."""
        return self._groups.get(cksum, [])

    def insert(self, path, row):
        """Insert a path into the trie, with metadata in `row`"""
        components = [comp for comp in path.split('/') if comp]
        curr = _lookup_root_path_index(
            self.root_paths,
            components) or self.root

        new_nodes = [(curr, False)]

        for idx, name in enumerate(components):
            node = curr.children.get(name)
            if node is None:
                node = curr.append(name)
                self.nodes[id(node)] = node
                new_nodes.append((node, True))
            else:
                new_nodes.append((node, False))

            curr = node

        curr.make_leaf(row)

        self._groups[row[Column.CKSUM]].append(curr)
        self.max_depth = max(self.max_depth, curr.depth)
        return new_nodes

    def find(self, path):
        """Find a PathNode in the trie by its path"""
        curr = self.root
        for name in (comp for comp in path.split('/') if comp):
            curr = curr.children.get(name)
            if curr is None:
                return None

        return curr

    def resolve(self, iter_path):
        """Resolve a list of indices to a node from root.
        The list might be for example [0, 0, 2] to go
        two layers down on the left and select the third
        node on this layer.

        The root path is located at [0].

        This is similar to GtkTreePath by purpose.
        """
        curr = self.root

        # [0] refers to root, strip first index:
        for idx in iter_path[1:]:
            curr = curr.indices[idx]

        return curr

    def sort(self, column_id, reverse=False, root=None):
        """Sort the trie nodes by their value in at `column_id`.
        If reverse is True, biger values appear first.

        This implementation is a generator that yields
        each visited node after sorting and a mapping from
        the new position to their old position
        """
        # Do the actual sorting:
        root = root or self.root
        children = deque(root.children.values())

        root.indices = sorted(
            children,
            key=lambda node: node[column_id],
            reverse=reverse
        )

        # Remember what changed:
        old_indices = []
        for idx, child in enumerate(root.indices):
            old_indices.append(child.idx)
            child.idx = idx

        yield root, old_indices

        # Propagate to children:
        for child in children:
            yield from self.sort(column_id, reverse, child)

    def has_leaves(self):
        """Check if this trie has any leaf nodes.
        This might be different from nodes that are inserted as sub root nodes,
        or as intermediate directories.
        """
        for node in self:
            if node.is_leaf:
                return True

        return False


def make_iter(node):
    """Make a GtkTreeIter, suitable for our PathTreeModel."""
    iter_ = Gtk.TreeIter()
    iter_.stamp = PATH_MODEL_STAMP
    iter_.user_data = id(node)
    return iter_


class PathTreeModel(GObject.GObject, Gtk.TreeModel, Gtk.TreeSortable):
    """Pack lint nodes into a tree structure compatible with Gtk

    This implements Gtk.TreeModel and can be therefore used as backend of
    Gtk.TreeView. It is backed up by a PathTrie as internal datastructure.

    Technical note:
        PyGObject prohibits storing pointers in Gtk.TreeIter.user_data.
        Therefore we store a hashtable with the id of our nodes and set the id
        as user_data as workaround. (see make_iter())
    """

    def __init__(self, paths):
        super(PathTreeModel, self).__init__()

        # Actual data storage:
        self.paths = paths
        self.trie = PathTrie(paths)

        # Manually insert the root path (we need one for working normally):
        path = Gtk.TreePath.new_from_indices([0])
        self.row_inserted(path, make_iter(self.trie.root))

        # Pack of files that are inserted to the trie at once.
        # This is a speed optimization to make the ui less blocking.
        self._file_pack, self._pack_timeout_id = [], None

        # Search optimization:
        # When typing '.pyo' after searching for '.py' we
        # can just filter the previous results.
        self._partial_model, self._last_query = None, None

        # Last column we sorted by or None (needed for GtkTreeSortable)
        self._sort_last_id = None
        self._sort_last_order = Gtk.SortType.DESCENDING

        # Set of nodes that need to get updated periodically.
        # It would be expensive to do that on every insert
        # (causing one row redraw each), therefore we do it every second.
        self._intermediate_nodes = set()
        self._mtime_cache = set()
        GLib.timeout_add(1000, self._update_intermediate_nodes)

        # A node was updated from the outside, i.e. by another model,
        # or by directly modifiying a PathNode or PathTrie.
        self.trie.connect('node-updated', self.on_node_updated)

    def _update_intermediate_nodes(self):
        """Make sure the intermediate nodes get updated in a slow
        but sufficient interval.
        """
        self._intermediate_nodes.update(self.trie.sub_roots)
        self._intermediate_nodes.add(self.trie.root)

        for node in self._intermediate_nodes:
            indices = node.build_iter_path()
            path = Gtk.TreePath.new_from_indices(indices)

            if id(node.row) not in self._mtime_cache:
                self._update_mtime(node.build_path(), node.row)

            self.row_changed(path, make_iter(node))

        # Reset for next time:
        self._intermediate_nodes = set()
        return True

    ########################
    #   Append Machinery   #
    ########################

    def add_path(self, path, row, immediately=False):
        """Add a path, including metadata, to the model.

        If immediately is False, the path will be cached and
        added after a small timeout as performance optimization.
        """
        self._update_mtime(path, row)

        if immediately:
            # Add it it immediately.
            self._add_and_signal(path, row)
        else:
            # Defer the addition a bit more.
            self._file_pack.append((path, row))
            if self._pack_timeout_id is None:
                self._pack_timeout_id = GLib.timeout_add(
                    PATH_MODEL_TIMEOUT_MS, self._add_defer
                )

    def _update_mtime(self, path, row):
        """Update the mtime by reading it from disk."""
        if row[Column.MTIME] <= 0 and id(row) not in self._mtime_cache:
            # Attempt to read the mtime from file:
            try:
                row[Column.MTIME] = os.stat(path).st_mtime
            except OSError as err:
                LOGGER.debug('stat: %s', str(err))

            self._mtime_cache.add(id(row))

    def _add_and_signal(self, path, row):
        """Actually add the path and its metadata here.
        Also signal the GtkTreeView to update if necessary.
        """
        parents = self.trie.insert(path, row)

        for node, was_new in parents:
            indices = node.build_iter_path()
            path = Gtk.TreePath.new_from_indices(indices)

            if was_new:
                self.row_inserted(path, make_iter(node))

            self._intermediate_nodes.add(node)

    def _add_defer(self):
        """Add a pack of paths to the trie, max 500 at the same time."""
        LOGGER.info(
            'Adding pack: %d/%d',
            min(PATH_MODEL_CHUNK_SIZE, len(self._file_pack)),
            PATH_MODEL_CHUNK_SIZE
        )

        for path, row in self._file_pack[:PATH_MODEL_CHUNK_SIZE]:
            self._add_and_signal(path, row)

        self._file_pack = self._file_pack[PATH_MODEL_CHUNK_SIZE:]

        if self._file_pack:
            # If we had some leftovers we need to schedule a new _add_defer.
            self._pack_timeout_id = GLib.timeout_add(
                PATH_MODEL_TIMEOUT_MS, self._add_defer
            )
        else:
            # Finished for now, new paths might come later.
            self._pack_timeout_id = None

        return False

    def lookup_by_path(self, path):
        """Calls trie.find() to find the node attached to a path"""
        return self.trie.find(path)

    ##################################
    #     Filter Implementation      #
    ##################################

    def filter_model(self, term):
        """Filter the model (and thus update the view) by `query`.
        Instead of modifying the model, a new model is returned,
        which shows only contains the filtered nodes.

        This is surprisingly the most efficient method here, and comes with
        very little effort.  Small drawback: External code cannot rely on a
        single model
        """
        if len(term) < 2:
            LOGGER.debug('too short, showing full model')
            return self

        term = term.lower()
        partial_model = PathTreeModel(self.paths)
        partial_model._mtime_cache = self._mtime_cache

        query = Query.parse(term)
        if query is None:
            return

        # Find out which trie to filter.
        # If we had a search query with matching prefix before we can just
        # use the previous resulting model.
        base_trie = self.trie
        if self._partial_model and query.issubset(self._last_query):
            base_trie = self._partial_model.trie

        # Iterate over the trie; do not add unmatched.
        for node in base_trie:
            # For now we only search through leafs.
            if not node.is_leaf:
                continue

            if not query.matches(
                    node,
                    node[Column.SIZE],
                    node[Column.MTIME],
                    -node[Column.COUNT]
            ):
                continue

            # Do not copy the rows, just ref them.
            path = node.build_path()
            partial_model.add_path(path, node.row, True)

        # Remember for the next step:
        self._partial_model, self._last_query = partial_model, query
        return partial_model

    ###################################
    # PyGObject convinience interface #
    ###################################

    def set_value(self, iter_, column, value):
        """Set the value of a cell.

        PyGObject seems to expect a method with this name,
        when setting the value of a row.
        """
        node = self.trie.nodes[iter_.user_data]
        node.row[column] = value

        # Find out which path to update:
        indices = node.build_iter_path()
        path = Gtk.TreePath.new_from_indices(indices)

        # Signal our change:
        self.row_changed(path, iter_)

    def __len__(self):
        return len(self.trie)

    def on_node_updated(self, trie, node_id):
        """Called when a node was changed on the outside."""
        node = trie.lookup_node_id(node_id)

        # Mark for updating:
        if node is not None:
            self._intermediate_nodes.add(node)

    def mark_for_update(self, node):
        """Mark this node to be updated soon."""
        self._intermediate_nodes.add(node)

    ##################################
    # Tree Model Spec Implementation #
    ##################################

    def do_get_iter(self, path):
        """Returns a new TreeIter that points at path.

        The implementation returns a 2-tuple (bool, TreeIter|None).
        """
        node = self.trie.resolve(path.get_indices())

        if node is not None:
            return (True, make_iter(node))
        else:
            return (False, None)

    def _iter_move(self, iter_, offset):
        """Move iter_ by a certain offset."""
        node = self.trie.nodes[iter_.user_data]
        next_node = node.neighbor(offset)

        if next_node is None:
            return (False, None)
        else:
            iter_.user_data = id(next_node)
            return (True, iter_)

    def do_iter_next(self, iter_):
        """Returns an iter pointing to the next row or None.

        The implementation returns a 2-tuple (bool, TreeIter|None).
        """
        return self._iter_move(iter_, +1)

    def do_iter_previous(self, iter_):
        """Returns an iter pointing to the previous row or None.

        The implementation returns a 2-tuple (bool, TreeIter|None).
        """
        return self._iter_move(iter_, -1)

    def do_iter_parent(self, child_iter):
        """Returns an iter pointing to the parent of child_iter or None."""
        node = self.trie.nodes[child_iter.user_data]
        if node.parent:
            return (True, make_iter(node.parent))
        else:
            return (False, None)

    def do_iter_has_child(self, iter_):
        """True if iter has children."""
        return len(self.trie.nodes[iter_.user_data].children) > 0

    def do_iter_n_children(self, iter_):
        """Returns the number of children of iter_"""
        if iter_ is None:
            return 0
        else:
            return len(self.trie.nodes[iter_.user_data].children)

    def do_iter_children(self, parent):
        """Return first child or (False|None)"""
        return self.do_iter_nth_child(parent, 0)

    def do_iter_nth_child(self, parent, nth):
        """Return iter that is set to the nth child of iter."""
        if parent is None:
            node = self.trie.root
        elif parent.user_data == 0:
            # Sometimes an invalid iterator lands here.
            # It has no user_data and an invalid stamp field.
            return (False, None)
        else:
            node = self.trie.nodes[parent.user_data]

        if nth < len(node.indices):
            return (True, make_iter(node.indices[nth]))

        return (False, None)

    def do_get_path(self, iter_):
        """Returns tree path references by iter."""
        node = self.trie.nodes[iter_.user_data]
        return Gtk.TreePath(reversed([parent.idx for parent in node.up()]))

    def do_get_value(self, iter_, column):
        """Returns the value for iter and column."""
        node = self.trie.nodes[iter_.user_data]
        return Column.TYPES[column](node[column])

    def do_get_n_columns(self):
        """Returns the number of columns."""
        return len(Column.TYPES)

    def do_get_column_type(self, column_idx):
        """Returns the type of the column."""
        return Column.TYPES[column_idx]

    def do_get_flags(self):
        """Returns the flags supported by this interface."""
        return Gtk.TreeModelFlags.ITERS_PERSIST

    ########################
    # Misc utility methods #
    ########################

    def iter_to_node(self, iter_):
        """Convert a GtkTreeIter to the related PathNode"""
        return self.trie.nodes.get(iter_.user_data)

    ###########################
    # GtkTreeSortable methods #
    ###########################

    def do_get_sort_column_id(self):
        """Return the column the model is sorted by.
        Returns (True if sorted, columnd id, order)
        """
        # Assume no initial sort order.
        if self._sort_last_id is None:
            id_ = Gtk.TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID
            return (False, id_, self._sort_last_order)
        else:
            return (True, self._sort_last_id, self._sort_last_order)

    def do_set_sort_column_id(self, id_, order):
        """Sort this model by the values in the column `id_` by `order`.
        The changes should be visible in the treeview immediately.
        """
        if id_ is Gtk.TREE_SORTABLE_DEFAULT_SORT_COLUMN_ID:
            return

        if id_ is Gtk.TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID:
            id_ = Column.PATH

        self._sort_last_id = id_
        self._sort_last_order = order

        # Indicate the sort column was changed *before* sorting.
        self.emit('sort-column-changed')

        # Do the actual sort:
        reverse = order is Gtk.SortType.DESCENDING
        for node, old_ind in self.trie.sort(id_, reverse):
            indices = node.build_iter_path()
            path = Gtk.TreePath.new_from_indices(indices)

            # Update the treeview:
            if old_ind:
                self.rows_reordered(path, make_iter(node), old_ind)

    def do_set_sort_func(self, id_, func):
        """Custom sort functions are hard to implement with tries."""
        raise NotImplementedError('Custom sort funcs are not supported.')

    def do_set_default_sort_func(self, id_, func):
        """Custom sort functions are hard to implement with tries."""
        raise NotImplementedError('Custom sort funcs are not supported.')

    def do_has_default_sort_func(self):
        """See above, not supported."""
        return False

    def sort(self, id_, order=Gtk.SortType.ASCENDING):
        """Convinience method for do_set_sort_column_id."""
        self.do_set_sort_column_id(id_, order)


def _create_column(title, id_, renderers, fixed_width=100):
    """Convinience method for creating a TreeView Column.
    Several renderers can be given with certain options.
    """
    column = Gtk.TreeViewColumn()
    column.set_title(title)
    column.set_resizable(True)
    column.set_sizing(Gtk.TreeViewColumnSizing.FIXED)
    column.set_fixed_width(fixed_width)

    if id_ is not None:
        column.set_sort_column_id(id_)

    for renderer, pack_end, expand, kwargs in renderers:
        renderer.set_alignment(1.0 if pack_end else 0.0, 0.5)
        pack_func = column.pack_end if pack_end else column.pack_start
        pack_func(renderer, expand)
        column.set_attributes(cell_renderer=renderer, **kwargs)
        column.set_expand(expand)

    return column


class PathTreeView(Gtk.TreeView):
    """A GtkTreeView that is readily configured for using PathTreeModel"""
    def __init__(self):
        Gtk.TreeView.__init__(self)

        # Enable separator lines:
        self.set_grid_lines(Gtk.TreeViewGridLines.NONE)
        self.set_enable_tree_lines(True)

        # Make selecting multiple rows possible:
        self.get_selection().set_mode(Gtk.SelectionMode.MULTIPLE)

        # Small spedup:
        self.set_fixed_height_mode(True)

        # Enable querying of the tooltip column:
        self.set_tooltip_column(Column.TOOLTIP)

        # Configure the column rendering:
        self.append_column(_create_column(
            'Path', Column.PATH, [
                (CellRendererLint(), False, False, dict(tag=Column.TAG)),
                (Gtk.CellRendererText(), False, True, dict(text=Column.PATH)),
            ],
            240
        ))
        self.append_column(_create_column(
            'Size', Column.SIZE,
            [(CellRendererSize(), True, False, dict(size=Column.SIZE))],
            70
        ))
        self.append_column(_create_column(
            'Count', Column.COUNT,
            [(CellRendererCount(), True, False, dict(count=Column.COUNT))],
            90
        ))
        self.append_column(
            _create_column(
                'Changed', Column.MTIME, [
                    (CellRendererModifiedTime(), True, False, dict(
                        mtime=Column.MTIME))], 100))

        self.connect(
            'button-press-event',
            PathTreeView.on_button_press_event
        )

        # Shut up, pylint.
        self._menu = None

    def set_model(self, model):
        """Overwrite Gtk.TreeView.set_model, but expand sub root paths"""
        Gtk.TreeView.set_model(self, model)
        self.expand_all()

    def get_selected_nodes(self):
        """Extra convinience method for getting the currently selected nodes"""
        model, rows = self.get_selection().get_selected_rows()
        for tp_path in rows:
            node = model.trie.resolve(tp_path.get_indices())
            yield node

    def get_selected_node(self):
        """Return the first selected node or None."""
        try:
            return next(self.get_selected_nodes())
        except StopIteration:
            return None

    def on_button_press_event(self, event):
        """Callback handler only used for mouse clicks."""
        if event.button != 3:
            return

        menu = self.on_show_menu()
        if menu:
            menu.simple_popup(event)

    #######################
    # MENU ENTRY HANDLING #
    #######################

    def on_show_menu(self):
        """Called during the button-press-event to show the actual menu."""
        # HACK: bind to self, since the ref would get lost.
        self._menu = PopupMenu()
        self._menu.simple_add('Toggle all', self.on_toggle_all)
        self._menu.simple_add('Toggle selected', self.on_toggle_selected)
        self._menu.simple_add_separator()
        self._menu.simple_add('Expand all', self.on_expand_all)
        self._menu.simple_add('Collapse all', self.on_collapse_all)
        self._menu.simple_add_separator()
        self._menu.simple_add('Open item', self.on_open_folder)
        self._menu.simple_add(
            'Copy path to clipboard',
            self.on_copy_to_clipboard
        )
        return self._menu

    def on_open_folder(self, _):
        """Open the selected item or folder via xdg-open."""
        node = self.get_selected_node()
        if node is None:
            return

        try:
            LOGGER.info('Calling xdg-open %s', node.build_path())
            Gio.Subprocess.new(
                ['xdg-open', node.build_path()], 0
            )
        except GLib.Error as err:
            LOGGER.exception('Could not open directory via xdg-open')

    def on_copy_to_clipboard(self, _):
        """Copy the currently selected full path to the clipboard."""
        node = self.get_selected_node()
        if node is None:
            return

        path = node.build_path()
        clipboard = Gtk.Clipboard.get_default(Gdk.Display.get_default())
        clipboard.set_text(path, len(path))

    def _toggle_tag_state(self, node_iter):
        """Iterate over `node_iter` and toggle the `tag` state."""
        model = self.get_model()
        for node in node_iter:
            current, new = node[Column.TAG], NodeState.NONE

            if current is NodeState.DUPLICATE:
                new = NodeState.ORIGINAL
            elif current is NodeState.ORIGINAL:
                new = NodeState.DUPLICATE

            self.update_node(node, Column.TAG, new)

    def on_toggle_all(self, _):
        """Toggle all nodes in the current visible model."""
        model = self.get_model()
        self._toggle_tag_state(model.trie)

    def on_toggle_selected(self, _):
        """Toggle all selected nodes in the current visible model."""
        nodes = list(self.get_selected_nodes())
        self._toggle_tag_state(nodes)

        # Note: this is the *full* trie.
        model = self.get_model()
        trie = model.trie

        for node in nodes:
            # Json documents with all related twins:
            group = trie.group(node[Column.CKSUM])
            if not group:
                continue

            # List of PathNodes which are twins:
            has_original = False
            for twin_node in group:
                if twin_node[Column.TAG] is NodeState.ORIGINAL:
                    has_original = True
                    break

            # All good:
            if has_original:
                continue

            # No original in group, set first twin as original.
            for twin_node in group:
                if twin_node is not node:
                    self.update_node(twin_node, Column.TAG, NodeState.ORIGINAL)
                    break

    def on_expand_all(self, _):
        """Just expand everything in the tree."""
        self.expand_all()

    def on_collapse_all(self, _):
        """Just collpase everything in the tree."""
        self.collapse_all()

    def set_twin(self, view):
        self.twin_view = view

    def update_node(self, node, column_id, value):
        main_model, twin_model = self.get_model(), self.twin_view.get_model()
        for model in [m for m in (main_model, twin_model) if m]:
            model.trie.update_node(node, column_id, value)

            for child in model.trie.group(node[Column.CKSUM]):
                model.mark_for_update(child)


if __name__ == '__main__':
    def main():
        """Show a window with the dupe-contents of a user specified path."""
        import sys

        model = PathTreeModel(sys.argv[1:])
        for arg_path in sys.argv[1:]:
            model.add_path(
                arg_path, Column.make_row({'mtime': time.time(), 'size': 0}))

        from shredder.runner import Runner
        settings = Gio.Settings.new('org.gnome.Shredder')

        runner = Runner(settings, sys.argv[1:], [])
        runner.connect(
            'lint-added',
            lambda *_: model.add_path(
                runner.element['path'],
                Column.make_row(
                    runner.element)))

        runner.connect(
            'process-finished',
            lambda _, msg: print(
                'Status:',
                msg))
        runner.run()

        view = PathTreeView()
        view.set_model(model)

        runner.connect(
            'process-finished',
            lambda _, msg: GLib.timeout_add(
                500,
                view.expand_all))

        runner.connect(
            'process-finished',
            lambda *_: GLib.timeout_add(
                500,
                lambda: model.sort(Column.SIZE)
            )
        )
        runner.connect(
            'process-finished',
            lambda *_: GLib.timeout_add(
                600,
                lambda: print(model.trie)
            )
        )

        def _search_changed(entry):
            view.set_model(model.filter_model(entry.get_text()))
            view.expand_all()

        entry = Gtk.SearchEntry()
        entry.connect('search-changed', _search_changed)

        scw = Gtk.ScrolledWindow()
        scw.add(view)

        win = Gtk.Window()
        win.set_default_size(640, 480)
        win.connect('destroy', Gtk.main_quit)

        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        box.pack_start(scw, True, True, 0)
        box.pack_start(entry, False, False, 0)

        win.add(box)
        win.show_all()

        Gtk.main()

    main()



#!/usr/bin/env python3
# encoding: utf-8
from nose import with_setup
from tests.utils import *


@with_setup(usual_setup_func, usual_teardown_func)
def test_simple():
    create_file('1234567890', 'a10')
    create_file('x23456789x', 'b10')

    head, *data, footer = run_rmlint_pedantic('-D')
    assert len(data) == 0

    for suffix in ['-Q .9 -q .1', '-Q 9 -q .1', '-Q .9 -q 1', '-Q 90% -q 10%']:
        head, *data, footer = run_rmlint('-D --rank-by a ' + suffix)
        assert data[0]['path'].endswith('a10')
        assert data[1]['path'].endswith('b10')
        assert len(data) == 2


@with_setup(usual_setup_func, usual_teardown_func)
def test_almost_empty():
    create_file('x', 'a1')
    create_file('x', 'b1')

    head, *data, footer = run_rmlint('-D -Q .5')
    assert len(data) == 0
    assert footer['total_files'] == 0

    head, *data, footer = run_rmlint('-D -q 1')
    assert len(data) == 0
    assert footer['total_files'] == 0


@with_setup(usual_setup_func, usual_teardown_func)
def test_absolute():
    data1 = ['x'] * 2048
    data2 = ['x'] * 2048
    data2[1023] = 'y'

    create_file(''.join(data1), 'a')
    create_file(''.join(data2), 'b')

    head, *data, footer = run_rmlint('-D')
    assert len(data) == 0

    head, *data, footer = run_rmlint_pedantic('-D -q 1kb -S a')
    assert data[0]['path'].endswith('a')
    assert data[1]['path'].endswith('b')
    assert len(data) == 2

    head, *data, footer = run_rmlint_pedantic('-D -q 1024 -S a')
    assert data[0]['path'].endswith('a')
    assert data[1]['path'].endswith('b')
    assert len(data) == 2

    head, *data, footer = run_rmlint_pedantic('-D -q 1023 -S a')
    assert len(data) == 0


@with_setup(usual_setup_func, usual_teardown_func)
def test_clamped_to_empty():
    create_file('x', 'empties/a')
    create_file('x', 'empties/b')

    head, *data, footer = run_rmlint('-q 0 -Q 0')
    assert len(data) == 0
    assert footer['total_files'] == 0

    head, *data, footer = run_rmlint('-q 1 -Q 1')
    assert len(data) == 0
    assert footer['total_files'] == 0

    head, *data, footer = run_rmlint('-q 10 -Q 1')
    assert len(data) == 0
    assert footer['total_files'] == 0

    # Just to check the test actually works normally:
    head, *data, footer = run_rmlint('-q 0 -Q 10')
    assert len(data) == 2
    assert footer['total_files'] == 2



#!/usr/bin/env python3
# encoding: utf-8
def teardown():
    from .utils import usual_teardown_func

    try:
        usual_teardown_func()
    except OSError:
        pass


def setup():
    from .utils import usual_setup_func
    usual_setup_func()



#!/usr/bin/env python3
# encoding: utf-8
#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# libmunin api doc documentation build configuration file, created by
# sphinx-quickstart on Mon Oct 21 16:42:48 2013.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys
import os

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath('./'))

try:
    import sphinx_bootstrap_theme

    # Activate the theme.
    templates_path = ['_templates']
    html_theme = 'bootstrap'
    html_theme_path = sphinx_bootstrap_theme.get_html_theme_path()
except ImportError:
    print('Cannot import `sphinx_bootstrap_theme`; falling back to `nature`.')
    print('^ This is no error, will cause only slightly different html output.')
    html_theme = 'nature'

# -- General configuration -----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.mathjax',
    'sphinx.ext.ifconfig',
    'sphinx.ext.graphviz',
    'sphinx.ext.viewcode'
]

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
#source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

todo_include_todos = True

# General information about the project.
project = 'rmlint documentation'
copyright = '2014-2019, Christopher Pahl & Daniel Thomas'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
#today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build', '_themes/armstrong/README.rst']

# The reST default role (used for this markup: `text`) to use for all documents.
#default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
#add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
show_authors = True

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "default"

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []


# -- Options for HTML output ---------------------------------------------------

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.

# (Optional) Logo. Should be small enough to fit the navbar (ideally 24x24).
# Path should be relative to the ``_static`` files directory.
html_logo = "logo.png"

# Theme options are theme-specific and customize the look and feel of a
# theme further.
if html_theme == 'bootstrap':
    html_theme_options = {
        # Navigation bar title. (Default: ``project`` value)
        'navbar_title': " ",

        # Tab name for entire site. (Default: "Site")
        'navbar_site_name': "Topics",

        # A list of tuples containing pages or urls to link to.
        # Valid tuples should be in the following forms:
        #    (name, page)                 # a link to a page
        #    (name, "/aa/bb", 1)          # a link to an arbitrary relative url
        #    (name, "http://example.com", True) # arbitrary absolute url
        # Note the "1" or "True" value above as the third argument to indicate
        # an arbitrary url.
        'navbar_links': [
            ("GitHub", "https://github.com/sahib/rmlint", True),
            ("Travis", "https://travis-ci.org/sahib/rmlint", True)
        ],

        # Render the next and previous page links in navbar. (Default: true)
        'navbar_sidebarrel': False,

        # Render the current pages TOC in the navbar. (Default: true)
        'navbar_pagenav': True,

        # Tab name for the current pages TOC. (Default: "Page")
        'navbar_pagenav_name': "Chapters",

        # Global TOC depth for "site" navbar tab. (Default: 1)
        # Switching to -1 shows all levels.
        'globaltoc_depth': 2,

        # Include hidden TOCs in Site navbar?
        #
        # Note: If this is "false", you cannot have mixed ``:hidden:`` and
        # non-hidden ``toctree`` directives in the same page, or else the build
        # will break.
        #
        # Values: "true" (default) or "false"
        'globaltoc_includehidden': "true",

        # HTML navbar class (Default: "navbar") to attach to <div> element.
        # For black navbar, do "navbar navbar-inverse"
        'navbar_class': "navbar",

        # Fix navigation bar to top of page?
        # Values: "true" (default) or "false"
        'navbar_fixed_top': "true",

        # Location of link to source.
        # Options are "nav" (default), "footer" or anything else to exclude.
        'source_link_position': "footer",

        # Bootswatch (http://bootswatch.com/) theme.
        #
        # Options are nothing with "" (default) or the name of a valid theme
        # such as "amelia" or "cosmo".
        # 'bootswatch_theme': "simplex",
        # 'bootswatch_theme': "journal",
        # 'bootswatch_theme': "readable",
        #'bootswatch_theme': "sandstone",
        'bootswatch_theme': "readable",

        # Choose Bootstrap version.
        # Values: "3" (default) or "2" (in quotes)
        'bootstrap_version': "3",
    }

# Add any paths that contain custom themes here, relative to this directory.

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".

try:
    with open('../.version', 'r') as handle:
        version_string = handle.read().strip()
except OSError:
    version_string = '2.x.x'

html_title = 'rmlint ({v}) documentation'.format(v=version_string)

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
github_fork = 'sahib'

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = '_static/logo.ico'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = False
# smart_quotes = True

# Custom sidebar templates, maps document names to template names.
html_default_sidebars = ['localtoc.html', 'sourcelink.html', 'searchbox.html']
html_sidebars = {
    'tutorial': html_default_sidebars,
    'developers': html_default_sidebars,
    'install': html_default_sidebars,
    'translators': html_default_sidebars
}

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
html_domain_indices = True

# If false, no index is generated.
html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = False

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = 'rmlint'


# -- Options for LaTeX output --------------------------------------------------

latex_elements = {
# The paper size ('letterpaper' or 'a4paper').
#'papersize': 'letterpaper',

# The font size ('10pt', '11pt' or '12pt').
#'pointsize': '10pt',

# Additional stuff for the LaTeX preamble.
#'preamble': '',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
# latex_documents = [
#   ('index', 'libmuninapidoc.tex', 'libmunin api doc Documentation',
#    'Christopher Pahl', 'manual'),
# ]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
#latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
#latex_use_parts = False

# If true, show page references after internal links.
#latex_show_pagerefs = False

# If true, show URL addresses after external links.
#latex_show_urls = False

# Documents to append as an appendix to all manuals.
#latex_appendices = []

# If false, no module index is generated.
#latex_domain_indices = True


# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('rmlint.1', 'rmlint', 'find duplicate files and other space waste efficiently',
     ['Christopher Pahl', 'Daniel Thomas'], 1)
]

# If true, show URL addresses after external links.
man_show_urls = False


# -- Options for Texinfo output ------------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
# texinfo_documents = [
#   ('index', 'libmuninapidoc', 'libmunin api doc Documentation',
#    'Christopher Pahl', 'libmuninapidoc', 'One line description of project.',
#    'Miscellaneous'),
# ]

# Documents to append as an appendix to all manuals.
#texinfo_appendices = []

# If false, no module index is generated.
#texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
#texinfo_show_urls = 'footnote'


# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    'python': ('http://python.readthedocs.org/en/latest/', None),
    'sphinx': ('http://sphinx.readthedocs.org/en/latest/', None)
}

# Make __init__ and Class.__doc__ visible:
autoclass_content = 'both'



#!/usr/bin/env python3
# encoding: utf-8
from nose import with_setup
from tests.utils import *

import subprocess

from parameterized import parameterized


def run_shell_script(shell, sh_path, *args):
    return subprocess.check_output([
            os.path.join("/bin", shell),
            sh_path,
        ] + list(args),
        shell=False
    ).decode("utf-8")


def filter_part_of_directory(data):
    return [e for e in data if e['type'] != 'part_of_directory']


@with_setup(usual_setup_func, usual_teardown_func)
@parameterized([("sh", ), ("bash", ), ("dash", )])
def test_basic(shell):
    create_file('xxx', 'a')
    create_file('xxx', 'b')

    create_file('yyy', 'dir_a/a')
    create_file('zzz', 'dir_a/b')

    create_file('zzz', 'dir_b/a')
    create_file('yyy', 'dir_b/b')

    create_file('', 'empty')

    create_file('aaa', 'aaa')
    create_link('aaa', 'link_aaa', symlink=True)
    os.remove(os.path.join(TESTDIR_NAME, 'aaa'))

    head, *data, footer = run_rmlint('-D -S a -o sh:{t}/rmlint.sh'.format(t=TESTDIR_NAME))
    data = filter_part_of_directory(data)
    # subprocess.call('cat ' + os.path.join(TESTDIR_NAME, 'rmlint.sh'), shell=True)

    assert footer['duplicate_sets'] == 3
    assert footer['total_lint_size'] == 9
    assert footer['total_files'] == 8 # + 1
    assert footer['duplicates'] == 3

    # Dry run first; check if it did not accidentally delete something.
    sh_path = os.path.join(TESTDIR_NAME, 'rmlint.sh')
    text = run_shell_script(shell, sh_path, "-dn")
    head, *data, footer = run_rmlint('-D -S a')
    data = filter_part_of_directory(data)
    assert footer['duplicate_sets'] == 3
    assert footer['total_lint_size'] == 9
    assert footer['total_files'] == 9
    assert footer['duplicates'] == 3

    text = run_shell_script(shell, sh_path, "-d")
    head, *data, footer = run_rmlint('-D -S a')
    data = filter_part_of_directory(data)

    assert footer['duplicate_sets'] == 0
    assert footer['total_lint_size'] == 0
    assert footer['total_files'] == 3
    assert footer['duplicates'] == 0

    assert '/dir_a' in text
    assert '/a' in text


@parameterized([("sh", ), ("bash", ), ("dash", )])
@with_setup(usual_setup_func, usual_teardown_func)
def test_paranoia(shell):
    create_file('xxx', 'a')
    create_file('xxx', 'b')
    create_file('xxx', 'c')
    create_link('a', 'hardlink_a', symlink=False)

    head, *data, footer = run_rmlint(
        '-S a -o sh:{t}/rmlint.sh'.format(t=TESTDIR_NAME)
    )

    assert footer['duplicate_sets'] == 1
    assert footer['total_lint_size'] == 6
    assert footer['total_files'] == 4 # + 1
    assert footer['duplicates'] == 3

    # Modify c after running rmlint:
    with open(os.path.join(TESTDIR_NAME, 'c'), 'w') as handle:
        handle.write('xxxx')

    sh_script = os.path.join(TESTDIR_NAME, 'rmlint.sh')
    text = run_shell_script(shell, sh_script, '-d', '-p', '-x')

    # Change back 'c':
    with open(os.path.join(TESTDIR_NAME, 'c'), 'w') as handle:
        handle.write('xxx')

    head, *data, footer = run_rmlint(
        '-S a -o sh:{t}/rmlint.sh'.format(t=TESTDIR_NAME)
    )

    assert footer['duplicate_sets'] == 1
    assert footer['total_lint_size'] == 3
    assert footer['total_files'] == 2 # +1
    assert footer['duplicates'] == 1

    assert 'files no longer identical' in text

    # Remove original:
    os.remove(os.path.join(TESTDIR_NAME, 'a'))

    text = run_shell_script(shell, sh_script, '-d', '-p')
    head, *data, footer = run_rmlint('-S a -o sh:{t}/rmlint.sh'.format(t=TESTDIR_NAME))

    assert 'original has disappeared' in text

    assert footer['duplicate_sets'] == 0
    assert footer['total_lint_size'] == 0
    assert footer['total_files'] == 1 # + 1
    assert footer['duplicates'] == 0


@with_setup(usual_setup_func, usual_teardown_func)
def test_anon_pipe():
    create_file('xxx', 'long-dummy-file-1')
    create_file('xxx', 'long-dummy-file-2')

    data = run_rmlint(
        "-o sh:>(cat)",
        force_no_pendantic=True,
        directly_return_output=True,
        use_shell=True
    )

    assert b'/long-dummy-file-1' in data
    assert b'/long-dummy-file-2' in data


@parameterized([("sh", ), ("bash", ), ("dash", )])
@with_setup(usual_setup_func, usual_teardown_func)
def test_hardlink_duplicate_directories(shell):
    create_file('xxx', 'dir_a/x')
    create_file('xxx', 'dir_b/x')

    sh_path = os.path.join(TESTDIR_NAME, "result.sh")
    header, *data, footer = run_rmlint(
        "-D -S a -c sh:hardlink -o sh:{}".format(sh_path),
    )
    data = filter_part_of_directory(data)
    assert len(data) == 2
    assert data[0]["path"].endswith("dir_a")
    assert data[1]["path"].endswith("dir_b")

    run_shell_script(shell, sh_path, "-d")

    full_dupe_a = os.path.join(TESTDIR_NAME, "dir_a/x")
    full_dupe_b = os.path.join(TESTDIR_NAME, "dir_b/x")
    assert os.stat(full_dupe_a).st_ino == os.stat(full_dupe_b).st_ino


def _check_if_empty_dirs_deleted(shell, inverse_order, sh_path, data):
    run_shell_script(shell, sh_path, "-dc")

    if inverse_order:
        assert not os.path.exists(data[1]["path"])
        assert not os.path.exists(os.path.join(TESTDIR_NAME, "deep/a"))
        assert os.path.exists(data[0]["path"])
    else:
        assert os.path.exists(data[0]["path"])
        assert not os.path.exists(data[1]["path"])


@parameterized([
    ("sh", False), ("bash", False), ("dash", False),
    ("sh", True), ("bash", True), ("dash", True)
])
@with_setup(usual_setup_func, usual_teardown_func)
def test_remove_empty_dirs(shell, inverse_order):
    create_file('xxx', 'deep/a/b/c/d/e/1')
    create_file('xxx', 'deep/x/2')

    sh_path = os.path.join(TESTDIR_NAME, "result.sh")
    header, *data, footer = run_rmlint(
        "-S {} -o sh:{}".format(
            "A" if inverse_order else "a",
            sh_path
        ),
    )

    assert len(data) == 2

    if inverse_order:
        assert data[0]["path"].endswith("x/2")
        assert data[0]["is_original"] is True
        assert data[1]["path"].endswith("e/1")
        assert data[1]["is_original"] is False
    else:
        assert data[0]["path"].endswith("e/1")
        assert data[0]["is_original"] is True
        assert data[1]["path"].endswith("x/2")
        assert data[1]["is_original"] is False

    _check_if_empty_dirs_deleted(shell, inverse_order, sh_path, data)


@parameterized([
    ("sh", False), ("bash", False), ("dash", False),
    ("sh", True), ("bash", True), ("dash", True)
])
@with_setup(usual_setup_func, usual_teardown_func)
def test_remove_empty_dirs_with_dupe_dirs(shell, inverse_order):
    create_file('xxx', 'deep/a/b/c/d/e/1')
    create_file('xxx', 'deep/x/1')

    sh_path = os.path.join(TESTDIR_NAME, "result.sh")
    header, *data, footer = run_rmlint(
        "-S {} -Dj -o sh:{}".format(
            "A" if inverse_order else "a",
            sh_path
        ),
    )
    data = filter_part_of_directory(data)

    assert len(data) == 2

    if inverse_order:
        assert data[0]["path"].endswith("x")
        assert data[0]["is_original"] is True
        assert data[1]["path"].endswith("e")
        assert data[1]["is_original"] is False
    else:
        assert data[0]["path"].endswith("e")
        assert data[0]["is_original"] is True
        assert data[1]["path"].endswith("x")
        assert data[1]["is_original"] is False

    _check_if_empty_dirs_deleted(shell, inverse_order, sh_path, data)

@with_setup(usual_setup_func, usual_teardown_func)
@parameterized([("sh", ), ("bash", ), ("dash", )])
def test_cleanup_emptydirs(shell):
    create_file('xxx', 'dir1/a')

    # create some ugly dir names
    names = [ 'escape me [please?]', '上野洋子, 吉野裕司, 浅井裕子 & 河越重義', '天谷大輔', 'Аркона',
            'let\'s nest',
            'let\'s nest/a level',
            'let\'s nest/a level/[or two]' ]
    for dirname in names:
        create_file('xxx', '{}/b'.format(dirname))

    head, *data, footer = run_rmlint('-S a -T df -o sh:{t}/rmlint.sh'.format(t=TESTDIR_NAME))

    assert footer['duplicate_sets'] == 1
    assert footer['total_lint_size'] == 3 * len(names)
    assert footer['total_files'] == 1 + len(names)
    assert footer['duplicates'] == len(names)

    # run rmlint.sh with -c option (should clean up empty dirs after deleting 'b' files).
    sh_path = os.path.join(TESTDIR_NAME, 'rmlint.sh')
    text = run_shell_script(shell, sh_path, "-dc")

    assert os.path.exists(os.path.join(TESTDIR_NAME, 'dir1/a'))

    for dirname in names:
        assert (not os.path.exists(os.path.join(TESTDIR_NAME, dirname)))



@with_setup(usual_setup_func, usual_teardown_func)
@parameterized([("sh", ), ("bash", ), ("dash", )])
def test_keep_parent_timestamps(shell):
    create_file('xxx', 'dir/a')
    create_file('xxx', 'dir/b')

    dir_path = os.path.join(TESTDIR_NAME, 'dir')
    stat_before = os.stat(dir_path)

    head, *data, footer = run_rmlint('-S a -T df -o sh:{t}/rmlint.sh'.format(t=TESTDIR_NAME))
    assert footer['duplicate_sets'] == 1
    assert footer['total_lint_size'] == 3
    assert footer['total_files'] == 2
    assert footer['duplicates'] == 1

    sh_path = os.path.join(TESTDIR_NAME, 'rmlint.sh')
    run_shell_script(shell, sh_path, "-dck")
    stat_after = os.stat(dir_path)

    assert stat_before.st_mtime == stat_after.st_mtime



#!/usr/bin/env python3
# encoding: utf-8
from nose import with_setup
from tests.utils import *


@with_setup(usual_setup_func, usual_teardown_func)
def test_km():
    # create some dupes with different paths, names and mtimes:
    create_file('xxx', 'stuff/a')
    create_file('yyy', 'stuff/b')
    create_file('yyy', 'stuff/b_dupe')
    create_file('xxx', 'morestuff/a_copy')
    create_file('yyy', 'morestuff/b_copy')
    create_file('xxx', 'backup/a_backup')
    create_file('xxx', 'backup/a_copy_backup')
    create_file('yyy+', 'backup/b_backup_changed')
    create_file('zzz', 'backup/d')
    create_file('zzz', 'backup/d_copy')

    # search path with backup folder tagged
    search_paths = TESTDIR_NAME + ' // ' + TESTDIR_NAME + '/backup'

    # 1. normal case - should find all dupes
    head, *data, footer = run_rmlint(search_paths, use_default_dir=False)
    assert len(data) == 4 + 2 + 3  # 4*'xxx' + 2*'zzz + 3*'yyy'
    assert footer['duplicates'] == (4 - 1) + (2 - 1) + (3 - 1)

    # 2. --keep-all-tagged case - should ignore 'zzz' pair since they are both tagged
    #                          should also treat 2 of the 'xxx' as originals since two are tagged
    head, *data, footer = run_rmlint('-k ' + search_paths, use_default_dir=False)
    assert len(data) == 4 + 3
    assert footer['duplicates'] == (4 - 2) + (3 - 1)
    for file_dict in data:
        if file_dict['type'] == 'duplicate_file' and file_dict['is_original'] == False:
            # check no tagged files marked as duplicates
            assert 'backup/' not in file_dict['path']

    # 3. --keep-all-untagged case - should ignore 'yyy' triple since they are all untagged
    #                          should also treat 2 of the 'xxx' as originals since two are untagged
    head, *data, footer = run_rmlint(' -K ' + search_paths, use_default_dir=False)
    assert len(data) == 4 + 2
    assert footer['duplicates'] == (4 - 2) + (2 - 1)
    for file_dict in data:
        if file_dict['type'] == 'duplicate_file' and file_dict['is_original'] == False:
            # check no untagged files marked as duplicates
            assert 'backup/' in file_dict['path']

    # 4. --must-match-tagged case - should ignore 'yyy' triple since they have no tagged copies
    head, *data, footer = run_rmlint(' -m ' + search_paths, use_default_dir=False)
    assert len(data) == 4 + 2
    assert footer['duplicates'] == (4 - 1) + (2 - 1)

    # 5. --must-match-untagged case - should ignore 'zzz' pair since they have no untagged copies
    head, *data, footer = run_rmlint(' -M ' + search_paths, use_default_dir=False)
    assert len(data) == 4 + 3
    assert footer['duplicates'] == (4 - 1) + (3 - 1)

    # 6. -km case:
    #     should ignore 'zzz' pair since they are both tagged
    #     should ignore 'yyy' triple since they have no tagged copies
    #     should also treat 2 of the 'xxx' as originals since two are tagged
    head, *data, footer = run_rmlint(' -km ' + search_paths, use_default_dir=False)
    assert len(data) == 4
    assert footer['duplicates'] == (4 - 2)
    for file_dict in data:
        if file_dict['type'] == 'duplicate_file' and file_dict['is_original'] == False:
            # check no tagged files marked as duplicates
            assert not 'backup/' in file_dict['path']

    # 7. -KM case:
    #     should ignore 'yyy' triple since they are all untagged
    #     should ignore 'zzz' pair since they have no untagged copies
    #     should also treat 2 of the 'xxx' as originals since two are untagged
    head, *data, footer = run_rmlint(' -KM ' + search_paths, use_default_dir=False)
    assert len(data) == 4
    assert footer['duplicates'] == (4 - 2)
    for file_dict in data:
        if file_dict['type'] == 'duplicate_file' and file_dict['is_original'] == False:
            # check no tagged files marked as duplicates
            assert 'backup/' in file_dict['path']



#!/usr/bin/env python3
# encoding: utf-8
from nose import with_setup
from tests.utils import *
import json


@with_setup(usual_setup_func, usual_teardown_func)
def test_stdin_read():
    path_a = create_file('1234', 'a')
    path_b = create_file('1234', 'b')
    path_c = create_file('1234', '.hidden')

    head, *data, footer = run_rmlint('-S a -D --partial-hidden', path_a, path_b, path_c)

    # Hidden is well, hidden due to --partial-hidden and -D.
    assert data[0]['path'].endswith('a')
    assert data[1]['path'].endswith('b')
    assert footer['total_lint_size'] == 8



#!/usr/bin/env python3
# encoding: utf-8
from nose import with_setup
from tests.utils import *


@with_setup(usual_setup_func, usual_teardown_func)
def test_simple():
    create_file('xxx', 'not_empty')
    create_file('', 'very_empty')
    head, *data, footer = run_rmlint('-T "none +ef"')

    assert footer['total_files'] == 2
    assert footer['duplicates'] == 0
    assert footer['total_lint_size'] == 0
    assert len(data) == 1
    assert data[0]['size'] == 0



#!/usr/bin/env python3
# encoding: utf-8
from nose import with_setup
from tests.utils import *

@with_setup(usual_setup_func, usual_teardown_func)
def test_negative():
    create_file('xxx', 'a.png')
    create_file('xxx', 'b.jpg')
    create_file('xxx', 'b')
    head, *data, footer = run_rmlint('-e')
    assert footer['total_files'] == 3
    assert footer['total_lint_size'] == 0
    assert footer['duplicates'] == 0


@with_setup(usual_setup_func, usual_teardown_func)
def test_positive():
    create_file('xxx', 'a.png')
    create_file('xxx', 'b.png')
    head, *data, footer = run_rmlint('-e')
    assert footer['total_files'] == 2
    assert footer['total_lint_size'] == 3
    assert footer['duplicates'] == 1



#!/usr/bin/env python3
# encoding: utf-8
from nose.plugins.attrib import attr
from nose import with_setup
from tests.utils import *

FILE_SIZE_KB = 10000
DIFFERENT_BYTES = 1
KBYTES_FROM_END = 10

@with_setup(usual_setup_func, usual_teardown_func)
def test_bigfiles():

    num_bytes = int(FILE_SIZE_KB * 1024)
    # create two identical files and a third one which differs near the end:
    create_file('x' * int(FILE_SIZE_KB * 1024), 'file1')

    create_file('x' * int(FILE_SIZE_KB * 1024), 'file2')

    create_file('x' * int((FILE_SIZE_KB - KBYTES_FROM_END )* 1024) +
                'y' * int(DIFFERENT_BYTES) +
                'x' * int(KBYTES_FROM_END * 1024 - DIFFERENT_BYTES),
                'file3')

    *_, footer = run_rmlint('')
    assert footer['duplicates'] == 1




#!/usr/bin/env python3
# encoding: utf-8
from nose import with_setup
from tests.utils import *
import os


SOURCE = '''
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>

int main(int argc, char *argv[]) {
    puts("Hello rmlint. Why were you executing this?");
    return EXIT_SUCCESS;
}
'''


def create_binary(path, stripped=False):
    path = path + '.stripped' if stripped else path + '.nonstripped'
    full_path = os.path.join(TESTDIR_NAME, path)

    command = 'echo \'{src}\' | cc -o {path} {option} -std=c99 -xc -'.format(
        src=SOURCE, path=full_path, option=('-s' if stripped else '-ggdb3')
    )
    subprocess.call(command, shell=True)


@with_setup(usual_setup_func, usual_teardown_func)
def test_negative():
    if has_feature('nonstripped') is False:
        return

    create_file(SOURCE, 'source.c')
    create_binary('source.c', stripped=True)
    head, *data, footer = run_rmlint('-T "none +nonstripped"')
    assert footer['total_files'] == 2
    assert footer['total_lint_size'] == 0
    assert len(data) == 0


@with_setup(usual_setup_func, usual_teardown_func)
def test_positive():
    if has_feature('nonstripped') is False:
        return

    create_file(SOURCE, 'source.c')
    create_binary('source.c', stripped=False)
    head, *data, footer = run_rmlint('-T "none +nonstripped"')
    assert footer['total_files'] == 2
    assert footer['total_lint_size'] == 0  # We cannot determine exact lint size.
    assert data[0]['type'] == 'nonstripped'



#!/usr/bin/env python3
# encoding: utf-8

from nose import with_setup
from nose.tools import make_decorator
from nose.plugins.skip import SkipTest
from contextlib import contextmanager
import psutil

from tests.utils import *

REFLINK_CAPABLE_FILESYSTEMS = {'btrfs', 'xfs', 'ocfs2'}

@contextmanager
def assert_exit_code(status_code):
    """
    Assert that the with block yields a subprocess.CalledProcessError
    with a certain return code. If nothing is thrown, status_code
    is required to be 0 to survive the test.
    """
    try:
        yield
    except subprocess.CalledProcessError as exc:
        assert exc.returncode == status_code
    else:
        # No exception? status_code should be fine.
        assert status_code == 0


def up(path):
    while path:
        yield path
        if path == "/":
            break
        path = os.path.dirname(path)

def is_on_reflink_fs(path):
    parts = psutil.disk_partitions(all=True)

    # iterate up from `path` until mountpoint found
    for up_path in up(path):
        for part in parts:
            if up_path == part.mountpoint:
                print("{0} is {1} mounted at {2}".format(path, part.fstype, part.mountpoint))
                return (part.fstype in REFLINK_CAPABLE_FILESYSTEMS)

    print("No mountpoint found for {0}", path)
    return False


# decorator for tests dependent on reflink-capable testdir
def needs_reflink_fs(test):
    def no_support(*args):
        raise SkipTest("btrfs not supported")

    def not_reflink_fs(*args):
        raise SkipTest("testdir is not on reflink-capable filesystem")

    if not has_feature('btrfs-support'):
        return make_decorator(test)(no_support)
    elif not is_on_reflink_fs(TESTDIR_NAME):
        return make_decorator(test)(not_reflink_fs)
    else:
        return test


@needs_reflink_fs
@with_setup(usual_setup_func, usual_teardown_func)
def test_equal_files():
    path_a = create_file('1234', 'a')
    path_b = create_file('1234', 'b')

    with assert_exit_code(0):
        run_rmlint(
            '--dedupe',
            path_a, path_b,
            use_default_dir=False,
            with_json=False,
            verbosity="")

    with assert_exit_code(0):
        run_rmlint(
            '--dedupe',
            path_a, '//', path_b,
            use_default_dir=False,
            with_json=False)


@needs_reflink_fs
@with_setup(usual_setup_func, usual_teardown_func)
def test_different_files():
    path_a = create_file('1234', 'a')
    path_b = create_file('4321', 'b')

    with assert_exit_code(1):
        run_rmlint(
            '--dedupe',
            path_a, path_b,
            use_default_dir=False,
            with_json=False,
            verbosity="")


@needs_reflink_fs
@with_setup(usual_setup_func, usual_teardown_func)
def test_bad_arguments():
    path_a = create_file('1234', 'a')
    path_b = create_file('1234', 'b')
    path_c = create_file('1234', 'c')
    for paths in [
            path_a,
            ' '.join((path_a, path_b, path_c)),
            ' '.join((path_a, path_a + ".nonexistent"))
    ]:
        with assert_exit_code(1):
            run_rmlint(
                '--dedupe',
                paths,
                use_default_dir=False,
                with_json=False,
                verbosity="")


@needs_reflink_fs
@with_setup(usual_setup_func, usual_teardown_func)
def test_directories():
    path_a = os.path.dirname(create_dirs('dir_a'))
    path_b = os.path.dirname(create_dirs('dir_b'))

    with assert_exit_code(1):
        run_rmlint(
            '--dedupe',
            path_a, path_b,
            use_default_dir=False,
            with_json=False,
            verbosity="")


@needs_reflink_fs
@with_setup(usual_setup_func, usual_teardown_func)
def test_dedupe_works():

    # test files need to be larger than btrfs node size to prevent inline extents
    path_a = create_file('1' * 100000, 'a')
    path_b = create_file('1' * 100000, 'b')

    # confirm that files are not reflinks
    with assert_exit_code(1):
        run_rmlint(
            '--is-reflink', path_a, path_b,
            use_default_dir=False,
            with_json=False,
            verbosity=""
        )

    # reflink our files
    with assert_exit_code(0):
        run_rmlint(
            '--dedupe', path_a, path_b,
            use_default_dir=False,
            with_json=False,
            verbosity=""
        )

    # confirm that they are now reflinks
    with assert_exit_code(0):
        run_rmlint(
            '--is-reflink', path_a, path_b,
            use_default_dir=False,
            with_json=False,
            verbosity=""
        )

# count the number of line in a file which start with patterns[]
def pattern_count(path, patterns):
    counts = [0] * len(patterns)
    with open(path, 'r') as f:
        for line in f:
            for i, pattern in enumerate(patterns):
                if line.startswith(pattern):
                    counts[i] += 1
    return counts


@needs_reflink_fs
@with_setup(usual_setup_func, usual_teardown_func)
def test_clone_handler():
    # test files need to be larger than btrfs node size to prevent inline extents
    path_a = create_file('1' * 100000, 'a')
    path_b = create_file('1' * 100000, 'b')

    sh_path = os.path.join(TESTDIR_NAME, 'rmlint.sh')

    # generate rmlint.sh and check that it correctly selects files for cloning
    with assert_exit_code(0):
        run_rmlint(
            '-S a -o sh:{p} -c sh:clone'.format(p=sh_path),
            path_a, path_b,
            use_default_dir=False,
            with_json=False
        )

    # parse output file for expected clone command
    counts = pattern_count(sh_path, ["clone '", "skip_reflink '"])
    assert counts[0] == 1
    assert counts[1] == 0

    # now reflink the two files and check again
    with assert_exit_code(0):
        run_rmlint(
            '--dedupe', path_a, path_b,
            use_default_dir=False,
            with_json=False,
            verbosity=""
        )
    with assert_exit_code(0):
        run_rmlint(
            '-S a -o sh:{p} -c sh:clone'.format(p=sh_path),
            path_a, path_b,
            use_default_dir=False,
            with_json=False
        )

    counts = pattern_count(sh_path, ["clone '", "skip_reflink '"])
    assert counts[0] == 0
    assert counts[1] == 1



#!/usr/bin/env python3
# encoding: utf-8
from nose.plugins.attrib import attr
from nose import with_setup
from tests.utils import *
import sys

NUMPAIRS = 1024+1

def branch_tree(current_path, remaining_depth):
    if (remaining_depth > 0):
        for i in range(2):
            next_path = current_path + "long" * 4 + str(i) + "/"
            create_dirs (next_path)
            branch_tree (next_path, remaining_depth - 1)
    else:
        for i in range(NUMPAIRS):
            create_file(str(i).zfill(1 + i), current_path + 'a' + str(i).zfill(7))
            create_file(str(i).zfill(1 + i), current_path + 'b' + str(i).zfill(7))
            create_link(current_path + 'a' + str(i).zfill(7), current_path + 'c' + str(i).zfill(7))
            create_link(current_path + 'a' + str(i).zfill(7), current_path + 'd' + str(i).zfill(7))
            create_link(current_path + 'a' + str(i).zfill(7), current_path + 'e' + str(i).zfill(7))
            create_link(current_path + 'a' + str(i).zfill(7), current_path + 'f' + str(i).zfill(7))
            create_link(current_path + 'b' + str(i).zfill(7), current_path + 'g' + str(i).zfill(7))
            create_link(current_path + 'b' + str(i).zfill(7), current_path + 'h' + str(i).zfill(7))


@attr('slow')
@with_setup(usual_setup_func, usual_teardown_func)
def test_manylongpathfiles():
    max_depth = 10 # will give 8M files total if NUMPAIRS = 1024+1
    branch_tree ("", max_depth)

    head, *data, footer = run_rmlint('-c json:no_body')
    assert footer['duplicates'] + footer['duplicate_sets'] == NUMPAIRS * 2 ** max_depth * 8



#!/usr/bin/env python3
# encoding: utf-8
from nose import with_setup
from tests.utils import *

import time


# Note: This test has the assumpption that it runs fast enough.
#       The threshold should be 1 second, so rather hard to reach.
#       This will not work with PEDANTIC=1 & run_rmlint_pedantic:
#       -n will update the timestamp after each run, causing
#       unexpected results.


def create_set(create_stamp, iso8601=False):
    # Basic case, should work always:
    create_file('xxx', 'a')
    create_file('xxx', 'b')

    arguments = '-S a'
    if iso8601:
        arguments += ' -c stamp:iso8601'

    head, *data, footer, stamp = run_rmlint_once(
        arguments, outputs=['stamp'] if create_stamp else ['pretty']
    )
    assert len(data) == 2

    create_file('xxx', 'c')
    warp_file_to_future('c', 2)


@with_setup(usual_setup_func, usual_teardown_func)
def test_simple():
    create_set(False)
    now = time.time()
    head, *data, footer = run_rmlint_once('-S a -N ' + str(time.time()))
    assert len(data) == 3

    # Fake a wait of two seconds a try an ISO8601.
    # Should return all 3 files, since c is newer and pulls in a + b.
    for offset, expect in [(1, 3), (8, 0)]:
        iso_time = time.strftime(
            "%Y-%m-%dT%H:%M:%S%z",
            time.gmtime(now + offset)
        )

        head, *data, footer = run_rmlint_once('-S a -N ' + iso_time)
        assert len(data) == expect


@with_setup(usual_setup_func, usual_teardown_func)
def test_stamp_file():
    create_set(True, False)

    # Wait 3 seconds, so the new stamp file (written by -n)
    # will contain a more recent timestamp, which will yield to 0 results
    # in the third run.
    time.sleep(3)

    head, *data, footer = run_rmlint_once('-S a -n ' + os.path.join(TESTDIR_NAME, '.stamp-0'))
    assert len(data) == 3

    head, *data, footer = run_rmlint_once('-S a -n ' + os.path.join(TESTDIR_NAME, '.stamp-0'))
    assert len(data) == 0


@with_setup(usual_setup_func, usual_teardown_func)
def test_stamp_file_iso8601():
    create_set(True, True)

    head, *data, footer = run_rmlint_once('-S a -n ' + os.path.join(TESTDIR_NAME, '.stamp-0'))
    assert len(data) == 3



#!/usr/bin/env python3
# encoding: utf-8
from nose import with_setup
from tests.utils import *

import subprocess


RMLINT_DUMMY_GROUP = '__rmlint_dummy_group'
RMLINT_DUMMY_USER = '__rmlint_dummy_user'


def exec_cmds(cmds):
    for cmd in cmds:
        fmt_cmd = cmd.format(
            u=RMLINT_DUMMY_USER,
            g=RMLINT_DUMMY_GROUP,
            t=TESTDIR_NAME
        )

        try:
            subprocess.check_call(fmt_cmd, shell=True, stderr=subprocess.PIPE)
        except subprocess.CalledProcessError as err:
            print(cmd, 'failed:', err)


@with_setup(usual_setup_func, usual_teardown_func)
def test_bad_ids():
    if not runs_as_root():
        return

    exec_cmds([
        'groupadd {g}',
        'useradd -M -N  {u}',
    ])

    try:
        create_file('x', '1_bad_uid')
        create_file('y', '2_bad_gid')
        create_file('z', '3_bad_gid_and_uid')

        exec_cmds([
            'chown {u} {t}/1_bad_uid',
            'chgrp {g} {t}/2_bad_gid',
            'chown {u}:{g} {t}/3_bad_gid_and_uid'
        ])
    finally:
        exec_cmds([
            'userdel -r {u}',
            'groupdel {g}'
        ])

    head, *data, footer = run_rmlint('-S a')

    x, y, z = data
    assert x['path'].endswith('1_bad_uid')
    assert y['path'].endswith('2_bad_gid')
    assert z['path'].endswith('3_bad_gid_and_uid')

    assert x['type'] == 'baduid'
    assert y['type'] == 'badgid'
    assert z['type'] == 'badugid'

    assert footer['total_files'] == 3



#!/usr/bin/env python3
# encoding: utf-8
from nose import with_setup
from tests.utils import *


@with_setup(usual_setup_func, usual_teardown_func)
def test_just_call_it():
    create_file('1234', 'a')
    create_file('1234', 'b')

    # This test is more or less here to make sure some util functions
    # are called from our tests. We don't test any results; basically
    # only if they fatally crash or create valgrind errors.
    # Also, you shouldn't see any output on the test run.
    run_rmlint(
        '-S a', outputs=['fdupes', 'stamp', 'progressbar', 'summary', 'pretty', 'py']
    )

    # Check if the -g option does weird things. (i.e. segfault)
    subprocess.check_output(['./rmlint', '-g', '-c', 'progressbar:ascii', TESTDIR_NAME])
    subprocess.check_output(['./rmlint', '-g', '-c', 'progressbar:fancy', TESTDIR_NAME])
    subprocess.check_output(['./rmlint', '-g',  '-O' , 'fdupes', TESTDIR_NAME])
    subprocess.check_output(['./rmlint', '-g', TESTDIR_NAME])

    for silly_option in ['-ppp', '-PPPP']:
        try:
            subprocess.check_output(['./rmlint', '-VVV', silly_option, TESTDIR_NAME])
        except subprocess.CalledProcessError:
            pass
        else:
            assert False



#!/usr/bin/env python3
# encoding: utf-8
from nose import with_setup
from tests.utils import *
import os


def create_bad_link(link_name):
    link_name = os.path.join(TESTDIR_NAME, link_name)
    fake_target = link_name + '.target'
    with open(fake_target, 'w') as h:
        h.write('xxx')

    try:
        os.symlink(fake_target, link_name)
    finally:
        os.remove(fake_target)


@with_setup(usual_setup_func, usual_teardown_func)
def test_basic():
    create_bad_link('imbad')

    for option in ('-f', '-F', '--see-symlinks'):
        head, *data, footer = run_rmlint(option)

        assert len(data) == 1
        assert data[0]['type'] == 'badlink'
        assert data[0]['path'].endswith('imbad')



#!/usr/bin/env python
# encoding: utf-8

"""Common constants."""

import gi

gi.require_version('Gtk', '3.0')
gi.require_version('Rsvg', '2.0')
gi.require_version('PangoCairo', '1.0')
gi.require_version('Polkit', '1.0')
gi.require_version('GtkSource', '3.0')

# Name of your application:
APP_TITLE = 'Shredder'

# One sentence description of the application:
APP_DESCRIPTION = 'Find & clean duplicate files'

# Use boxy old menus or new popovers?
APP_USE_TRADITIONAL_MENU = False


def run_gui():
    """Fully take over and run the gui code."""
    # Stdlib:
    import sys

    # Internal:
    from shredder.cmdline import parse_arguments
    from shredder.logger import create_logger

    ROOT_LOGGER = create_logger(None)
    OPTIONS = parse_arguments(ROOT_LOGGER)

    # Later import due to logging.
    from shredder.application import Application

    if OPTIONS:
        # Gtk will take over now.
        APP = Application(OPTIONS)
        ROOT_LOGGER.info('Starting up.')
        sys.exit(APP.run([sys.argv[0]]))



#!/usr/bin/env python3
# encoding: utf-8
from nose import with_setup
from tests.utils import *


@with_setup(usual_setup_func, usual_teardown_func)
def test_default():
    # --see-symlinks should be on by default.
    create_file('xxx', 'a/z')
    create_link('a/z', 'a/x', symlink=True)
    create_file('xxx', 'b/z')
    create_link('b/z', 'b/x', symlink=True)
    create_link('b/z', 'b/y', symlink=True)

    head, *data, footer = run_rmlint()
    assert {e["path"][len(TESTDIR_NAME):] for e in data} == {
        '/b/x',
        '/b/y',
        '/b/z',
        '/a/z',
    }

@with_setup(usual_setup_func, usual_teardown_func)
def test_merge_directories_with_ignored_symlinks():
    # Badlinks should not forbid finding duplicate directories
    # when being filtered out during traversing with -T dd,df.
    create_file('xxx', 'a/z')
    create_link('bogus', 'a/link', symlink=True)
    create_file('xxx', 'b/z')
    create_link('bogus', 'b/link', symlink=True)

    head, *data, footer = run_rmlint('-T df,dd')
    assert {e["path"][len(TESTDIR_NAME):] for e in data if e["type"] == "duplicate_dir"} == {
        '/a',
        '/b',
    }

@with_setup(usual_setup_func, usual_teardown_func)
def test_order():
    create_file('xxx', 'a/z')
    create_link('a/z', 'a/x', symlink=True)
    create_file('xxx', 'b/z')
    create_link('b/z', 'b/x', symlink=True)
    create_link('b/z', 'b/y', symlink=True)

    head, *data, footer = run_rmlint('-F')
    assert len(data) == 2
    assert data[0]['path'].endswith('z')
    assert data[0]['is_original']
    assert data[1]['path'].endswith('z')

    head, *data, footer = run_rmlint('-f -S a')
    assert sum(p['is_original'] for p in data) == 1

    assert len(data) == 2
    assert data[0]['path'].endswith('z')
    assert data[0]['is_original']

    assert data[1]['path'].endswith('z')
    assert not data[1]['is_original']

    head, *data, footer = run_rmlint('-@ -S a')
    assert sum(p['is_original'] for p in data) == 2
    assert len(data) == 4

    if data[0]['path'].endswith('x'):
        sym_idx, file_idx = 0, 2
    else:
        sym_idx, file_idx = 2, 0

    # Same symlink -> dupe
    assert data[sym_idx]['path'].endswith('x')
    assert data[sym_idx]['is_original']
    assert data[sym_idx + 1]['path'].endswith('y')

    # Same file
    assert data[file_idx]['path'].endswith('z')
    assert data[file_idx]['is_original']
    assert data[file_idx + 1]['path'].endswith('z')



#!/usr/bin/env python3
# encoding: utf-8
from nose import with_setup
from nose.plugins.attrib import attr
from tests.utils import *

import time

from itertools import permutations, combinations


PATHS = ['b_dir/', 'a_dir/', 'c_dir/']


def path_index(file_path):
    for idx, path in enumerate(PATHS):
        if path in file_path:
            return idx
    assert False


def path_depth(file_path):
    return len([c for c in file_path if c == '/'])


# dispatcher for comparison tests
def validate_order(data, tests):
    testfuncs = {
        's': lambda g: sum([p['size'] for p in g]),
        'a': lambda g: os.path.basename(g[0]['path']).lower(),
        'm': lambda g: g[0]['mtime'],
        'p': lambda g: path_index(g[0]['path']),
        'n': lambda g: len(g)
    }

    groups, group = [], []
    for point in data:
        if point['is_original']:
            group = []
            groups.append(group)

        group.append(point)

    assert len(groups) == 2

    for a, b in combinations(groups, 2):
        for test in tests:
            cmp_a, cmp_b = (testfuncs[test.lower()](e) for e in [a, b])

            # Equal? Try again.
            if cmp_a == cmp_b:
                continue

            a_comes_first = test.islower()

            if (cmp_a < cmp_b) and a_comes_first:
                break

            if (cmp_b < cmp_a) and not a_comes_first:
                break

            # Something's wrong.
            assert False


@with_setup(usual_setup_func, usual_teardown_func)
def test_replay_match_basename():
    create_file('xxx', 'test1/a')
    create_file('xxx', 'test1/b')
    create_file('xxx', 'test2/a')

    replay_path = '/tmp/replay.json'

    head, *data, footer = run_rmlint('-o json:{p}'.format(
        p=replay_path
    ))

    assert len(data) == 3

    head, *data, footer = run_rmlint('--replay {p}'.format(
        p=replay_path
    ))

    assert len(data) == 3

    head, *data, footer = run_rmlint('--replay {p} --match-basename'.format(
        p=replay_path
    ))

    assert len(data) == 2
    paths = set([p['path'] for p in data])
    assert os.path.join(TESTDIR_NAME, 'test1/a') in paths
    assert os.path.join(TESTDIR_NAME, 'test2/a') in paths

    head, *data, footer = run_rmlint('--replay {p} --unmatched-basename'.format(
        p=replay_path
    ))

    assert len(data) == 3


@with_setup(usual_setup_func, usual_teardown_func)
def test_replay_hidden():
    create_file('xxx', 'test/.a')
    create_file('xxx', 'test/.b')

    replay_path = '/tmp/replay.json'

    head, *data, footer = run_rmlint('--hidden -o json:{p}'.format(
        p=replay_path
    ))

    assert len(data) == 2

    head, *data, footer = run_rmlint('--replay {p}'.format(
        p=replay_path
    ))

    assert len(data) == 0

    head, *data, footer = run_rmlint('--replay {p} --hidden'.format(
        p=replay_path
    ))

    assert len(data) == 2


@with_setup(usual_setup_func, usual_teardown_func)
def test_replay_must_match_tagged():
    create_file('xxx', 'test_a/a')
    create_file('xxx', 'test_b/a')

    replay_path = '/tmp/replay.json'

    head, *data, footer = run_rmlint('-o json:{p}'.format(
        p=replay_path
    ))

    assert len(data) == 2

    head, *data, footer = run_rmlint('--replay {p} {b} // {a} -m'.format(
        p=replay_path,
        a=os.path.join(TESTDIR_NAME, 'test_a'),
        b=os.path.join(TESTDIR_NAME, 'test_b')
    ))

    paths = set([(p['path'], p['is_original']) for p in data])
    assert (os.path.join(TESTDIR_NAME, 'test_b/a'), False) in paths
    assert (os.path.join(TESTDIR_NAME, 'test_a/a'), True) in paths


@attr('slow')
@with_setup(usual_setup_func, usual_teardown_func)
def test_sorting():
    # create some dupes with different PATHS, names and mtimes:
    create_file('xxx', PATHS[0] + 'a')
    create_file('xxx', PATHS[1] + 'bb')
    create_file('xxx', PATHS[2] + 'ccc')

    # Make sure it takes some time to re-reun
    time.sleep(1.25)
    create_file('xxxx', PATHS[0] + 'A')
    create_file('xxxx', PATHS[1] + 'B')
    create_file('xxxx', PATHS[2] + 'C')
    create_file('xxxx', PATHS[2] + 'D')

    joiner = ' ' + TESTDIR_NAME + '/'
    search_paths = joiner + joiner.join(PATHS)

    # Leave out 'o' for now, since its not really worth testing.
    opts = 'sampn'
    all_opts = opts + opts.upper()

    combos = []
    is_legal_combo = lambda x: len(x) == len(set(x.lower()))

    # Limit to 3-tuple combinations. 4-5-tuple combinations (just do +1)
    # are possible if you have enough time (=> ~(10! / 2) tests).
    for n_terms in range(1, len(opts) - 1):
        combos += filter(
            is_legal_combo,
            (''.join(p) for p in permutations(all_opts, n_terms))
        )

    replay_path = '/tmp/replay.json'

    for combo in combos:
        combo_str = '-y ' + combo
        head, *data, footer = run_rmlint(
            combo_str + search_paths, '-o json:{p}'.format(p=replay_path),
            use_default_dir=False
        )
        assert len(data) == 7

        validate_order(data, combo)

        head, *data, footer = run_rmlint(
            combo_str + search_paths, ' --replay ' + replay_path,
            use_default_dir=False
        )

        assert len(data) == 7

        validate_order(data, combo)


@with_setup(usual_setup_func, usual_teardown_func)
def test_replay_no_dir():
    # Regression test for #305.
    # --replay did not replay anything when not specifying some path.
    # (The current working directory was not set in this case correctly)

    create_file('xxx', 'sub/a')
    create_file('xxx', 'sub/b')
    create_file('xxx', 'c')

    current_cwd = os.getcwd()

    try:
        os.chdir(TESTDIR_NAME)
        replay_path = '/tmp/replay.json'
        head, *data, footer = run_rmlint(
                '-o json:{p}'.format(p=replay_path),
                use_default_dir=False,
        )
        assert len(data) == 3

        head, *data, footer = run_rmlint(
                '--replay {}'.format(replay_path),
                use_default_dir=False,
        )
        assert len(data) == 3
    finally:
        os.chdir(current_cwd)


@with_setup(usual_setup_func, usual_teardown_func)
def test_replay_unicode_fuckup():
    names = '上野洋子, 吉野裕司, 浅井裕子 & 河越重義', '天谷大輔', 'Аркона'

    create_file('xxx', names[0])
    create_file('xxx', names[1])
    create_file('xxx', names[2])

    replay_path = '/tmp/replay.json'

    head, *data, footer = run_rmlint('-o json:{p}'.format(p=replay_path))
    assert len(data) == 3
    assert set([os.path.basename(e['path']) for e in data]) == set(names)

    head, *data, footer = run_rmlint('--replay {p}'.format(p=replay_path))
    assert len(data) == 3
    assert set([os.path.basename(e['path']) for e in data]) == set(names)


@with_setup(usual_setup_func, usual_teardown_func)
def test_replay_tagged_order():
    create_file('xxx', 'a/1')
    create_file('xxx', 'a/2')
    create_file('xxx', 'b/1')
    create_file('xxx', 'b/2')

    replay_path_a = '/tmp/replay-a.json'
    replay_path_b = '/tmp/replay-b.json'

    # Create replay-a.json
    head, *data, footer = run_rmlint(
        '-S a {r} -o json:{p}'.format(r=TESTDIR_NAME + '/a', p=replay_path_a),
        use_default_dir=False
    )

    assert len(data) == 2
    assert data[0]['path'].endswith('a/1')
    assert data[1]['path'].endswith('a/2')

    # Create replay-b.json
    head, *data, footer = run_rmlint(
        '{r} -o json:{p}'.format(r=TESTDIR_NAME + '/b', p=replay_path_b),
        use_default_dir=False
    )

    assert len(data) == 2
    assert data[0]['path'].endswith('b/1')
    assert data[1]['path'].endswith('b/2')

    # Check if b.json is preferred over a.json
    head, *data, footer = run_rmlint(
        '-S a --replay {a} // {b}'.format(a=replay_path_a, b=replay_path_b)
    )
    assert len(data) == 4
    assert [p['is_original'] for p in data] == [True, False, False, False]

    assert data[0]['path'].endswith('b/1')
    assert data[1]['path'].endswith('b/2')
    assert data[2]['path'].endswith('a/1')
    assert data[3]['path'].endswith('a/2')

    # Check if a.json is preferred over b.json
    head, *data, footer = run_rmlint(
        '-S a --replay {b} // {a}'.format(a=replay_path_a, b=replay_path_b)
    )

    assert len(data) == 4
    assert [p['is_original'] for p in data] == [True, False, False, False]

    assert data[0]['path'].endswith('a/1')
    assert data[1]['path'].endswith('a/2')
    assert data[2]['path'].endswith('b/1')
    assert data[3]['path'].endswith('b/2')

    head, *data, footer = run_rmlint(
        '-S a --replay {a} // {b} -k'.format(a=replay_path_a, b=replay_path_b)
    )

    assert len(data) == 4
    assert [p['is_original'] for p in data] == [True, True, False, False]

    assert data[0]['path'].endswith('b/1')
    assert data[1]['path'].endswith('b/2')
    assert data[2]['path'].endswith('a/1')
    assert data[3]['path'].endswith('a/2')


@with_setup(usual_setup_func, usual_teardown_func)
def test_replay_duplicate_directory_size():
    create_file('xxx', 'a/xxx')
    create_file('xxx', 'b/xxx')

    replay_path = '/tmp/replay.json'
    head, *data, footer = run_rmlint('-o json:{p} -S a'.format(p=replay_path))
    assert len(data) == 2

    head, *data, footer = run_rmlint('--replay {p}'.format(p=replay_path))
    assert len(data) == 2
    assert footer['total_lint_size'] == 3


def create_pack_and_unpack_scenario():
    x_files = [
        "a/sub1/x1",
        "a/sub1/x2",
        "a/sub2/x3",
        "a/sub2_copy/x3",
        "b/sub1/x1",
        "b/sub1/x2",
        "b/sub2/x3",
        "b/sub2_copy/x3",
    ]

    for path in x_files:
        create_file("x", path)

    create_file("special", "special")
    create_file("special", "a/special")
    create_file("special", "b/special")

    create_file("", "a/empty")
    create_file("", "b/empty")


def data_by_type(data):
    result = {}
    for entry in data:
        path = entry["path"][len(TESTDIR_NAME):]
        # TODO: bring back is_original.
        # result.setdefault(entry["type"], {})[path] = entry["is_original"]
        result.setdefault(entry["type"], {})[path] = False

    return result


EXPECTED_WITH_TREEMERGE = {
    'emptyfile': {
        '/a/empty': False,
        '/b/empty': False,
    },
    "part_of_directory": {
        '/b/sub1/x1'      : False,
        '/a/sub2/x3'      : False,
        '/b/sub2/x3'      : False,
        '/a/sub2_copy/x3' : False,
        '/b/sub2_copy/x3' : False,
        '/a/sub1/x2'      : False,
        '/a/sub1/x1'      : False,
        '/b/sub1/x2'      : False,
        '/a/special'      : False,
        '/b/special'      : False,
    },
    'duplicate_dir': {
        '/a'           : False,
        '/b'           : False,
        '/a/sub2'      : False,
        '/a/sub2_copy' : False,
    },
    'duplicate_file': {
        '/a/sub1/x1' : False,
        '/a/sub1/x2' : False,
        '/a/sub2/x3' : False,
        '/a/special' : False,
        '/special'   : False,
    },
}

EXPECTED_WITHOUT_TREEMERGE = {
    'emptyfile': {
        '/a/empty': False,
        '/b/empty': False,
    },
    'duplicate_file': {
        '/b/sub1/x1'      : False, #True,
        '/a/sub1/x1'      : False,
        '/a/sub1/x2'      : False,
        '/a/sub2/x3'      : False,
        '/a/sub2_copy/x3' : False,
        '/b/sub1/x2'      : False,
        '/b/sub2/x3'      : False,
        '/b/sub2_copy/x3' : False,
        '/a/special'      : False, #True
        '/b/special'      : False,
        '/special'        : False,
    }
}


@with_setup(usual_setup_func, usual_teardown_func)
def test_replay_pack_directories():
    create_pack_and_unpack_scenario()

    # Do a run without -D and pack it later during --replay.
    replay_path = '/tmp/replay.json'

    head, *data, footer = run_rmlint('-o json:{p} -S ahD'.format(p=replay_path))
    assert len(data) == 13
    assert data_by_type(data) == EXPECTED_WITHOUT_TREEMERGE

    # Do the run without any packing first (it should yield the same result)
    head, *data, footer = run_rmlint('--replay {p} -S ahD'.format(p=replay_path))
    assert len(data) == 13
    assert data_by_type(data) == EXPECTED_WITHOUT_TREEMERGE

    # Do the run with packing dupes into directories now:
    head, *data, footer = run_rmlint('--replay {p} -S ahD -D'.format(p=replay_path))
    assert len(data) == 21
    assert data_by_type(data) == EXPECTED_WITH_TREEMERGE


@with_setup(usual_setup_func, usual_teardown_func)
def test_replay_unpack_directories():
    create_pack_and_unpack_scenario()

    # Do a run with -D and pack it later during --replay.
    replay_path = '/tmp/replay.json'
    head, *data, footer = run_rmlint('-o json:{p} -S ahD -D'.format(p=replay_path))

    assert len(data) == 21
    assert data_by_type(data) == EXPECTED_WITH_TREEMERGE

    # Do a normal --replay run first without any unpacking:
    head, *data, footer = run_rmlint('--replay {p} -S ahD -D'.format(p=replay_path))
    assert len(data) == 21
    assert data_by_type(data) == EXPECTED_WITH_TREEMERGE

    # # Do the run with unpacking the directories:
    head, *data, footer = run_rmlint('--replay {p} -S ahD'.format(p=replay_path))
    assert len(data) == 23

    # NOTE: In this special case we should still carry around the
    # part_of_directory elements from previous runs.
    expected = EXPECTED_WITHOUT_TREEMERGE
    expected["part_of_directory"] = EXPECTED_WITH_TREEMERGE["part_of_directory"]

    assert data_by_type(data) == expected



#!/usr/bin/env python3
# encoding: utf-8
from nose import with_setup
from tests.utils import *


@with_setup(usual_setup_func, usual_teardown_func)
def test_simple():
    create_file('xxx', 'not_empty/a')
    create_file('', 'empty_but_with_file/a')
    create_dirs('really_empty')
    head, *data, footer = run_rmlint('-T "none +ed +ef" -S a')

    assert footer['total_files'] == 3
    assert footer['duplicates'] == 0
    assert footer['total_lint_size'] == 0
    assert len(data) == 2
    assert data[0]['size'] == 0
    assert data[0]['type'] == "emptydir"
    assert data[1]['size'] == 0
    assert data[1]['type'] == "emptyfile"


@with_setup(usual_setup_func, usual_teardown_func)
def test_deep():
    create_dirs('1/2/3/4/5')
    head, *data, footer = run_rmlint('-T "none +ed"')

    assert data[0]['path'].endswith('5')
    assert data[1]['path'].endswith('4')
    assert data[2]['path'].endswith('3')
    assert data[3]['path'].endswith('2')
    assert data[4]['path'].endswith('1')

    create_file('', '1/2/3/showstopper')
    head, *data, footer = run_rmlint('-T "none +ed"')

    assert data[0]['path'].endswith('5')
    assert data[1]['path'].endswith('4')


@with_setup(usual_setup_func, usual_teardown_func)
def test_hidden():
    create_file('xxx', 'not_empty/.hidden')
    head, *data, footer = run_rmlint('-T "none +ed"')

    assert footer['total_files'] == 0
    assert len(data) == 0

    head, *data, footer = run_rmlint('-T "none +ed" --hidden')
    assert footer['total_files'] == 1
    assert len(data) == 0



#!/usr/bin/env python3
# encoding: utf-8
import os
from nose import with_setup
from tests.utils import *


def filter_part_of_directory(data):
    return [e for e in data if e['type'] != 'part_of_directory']


@with_setup(usual_setup_func, usual_teardown_func)
def test_rankby_simple():
    create_file('x', 'ax')
    create_file('x', 'ay')
    create_file('yyy', 'bx')
    create_file('yyy', 'by')

    head, *data, foot = run_rmlint('--sort-by a -S a')
    paths = [os.path.basename(p['path']) for p in data]
    assert paths == ['ax', 'ay', 'bx', 'by']

    head, *data, foot = run_rmlint('--sort-by S -S A')
    paths = [os.path.basename(p['path']) for p in data]
    assert paths == ['by', 'bx', 'ay', 'ax']


@with_setup(usual_setup_func, usual_teardown_func)
def test_rankby_dirs():
    create_file('x', 'ax')
    create_file('x', 'ay')
    create_file('yyy', 'b/x')
    create_file('yyy', 'b/y')
    create_file('yyy', 'c/x')
    create_file('yyy', 'c/y')
    create_file('x' * 64, 'dx')
    create_file('x' * 64, 'dy')

    head, *data, foot = run_rmlint('--sort-by s -S a -D')
    data = filter_part_of_directory(data)
    paths = [os.path.basename(p['path']) for p in data]
    assert paths == ['b', 'c', 'ax', 'ay', 'x', 'y', 'dx', 'dy']



#!/usr/bin/env python3
# encoding: utf-8
import sys
import json
import hashlib

from collections import defaultdict


def sha1_of_file(path):
    sha = hashlib.sha1()
    with open(path, 'rb') as f:
        while True:
            block = f.read(2 ** 10)
            if not block:
                break

            sha.update(block)
        return sha.hexdigest()



if __name__ == '__main__':
    head, *data, footer = json.load(sys.stdin)

    dupe_groups = defaultdict(list)
    for point in data:
        dupe_groups[point['checksum']].append(point['path'])

    for checksum, group in dupe_groups.items():
        checksum_set = set()
        for path in group:
            sha1 = sha1_of_file(path)
            checksum_set.add(sha1)
            if len(checksum_set) > 1:
                print('The following path differs:')
                print('    {c}={p}'.format(p=path, c=sha1))
                print('It was in a group with this checksum')

                checksum_set.remove(sha1)
                print('    {c}={p}'.format(p=group, c=checksum_set.pop()))
                sys.exit(-1)

    print('Everything fine.')



#!/usr/bin/env python3
# encoding: utf-8
from nose import with_setup
from tests.utils import *


def filter_part_of_directory(data):
    return [e for e in data if e['type'] != 'part_of_directory']


@with_setup(usual_setup_func, usual_teardown_func)
def test_simple():
    create_file('xxx', '.a/1')
    create_file('xxx', '.b/1')
    create_file('xxx', '.1')

    head, *data, footer = run_rmlint('--hidden')

    assert footer['duplicates'] == 2
    assert footer['ignored_folders'] == 0
    assert footer['ignored_files'] == 0
    assert footer['duplicate_sets'] == 1


@with_setup(usual_setup_func, usual_teardown_func)
def test_hidden():
    create_file('xxx', '.a/1')
    create_file('xxx', '.b/1')
    create_file('xxx', '.1')
    head, *data, footer = run_rmlint('--no-hidden')

    assert footer['duplicates'] == 0
    assert footer['ignored_folders'] == 2
    assert footer['ignored_files'] == 3
    assert footer['duplicate_sets'] == 0


@with_setup(usual_setup_func, usual_teardown_func)
def test_explicit():
    create_file('xxx', '.a/1')
    create_file('xxx', '.a/2')
    head, *data, footer = run_rmlint('--no-hidden', dir_suffix='.a')

    assert footer['duplicates'] == 1
    assert footer['ignored_folders'] == 0
    assert footer['ignored_files'] == 0
    assert footer['duplicate_sets'] == 1


@with_setup(usual_setup_func, usual_teardown_func)
def test_partial_hidden():
    create_file('1', 'a/.hidden')
    create_file('1', 'b/.hidden')
    create_file('1', '.hidden')

    head, *data, footer = run_rmlint('--no-hidden')
    assert len(data) == 0

    head, *data, footer = run_rmlint('--partial-hidden')
    assert len(data) == 0

    head, *data, footer = run_rmlint('--hidden')
    assert len(data) == 3
    assert all(p['path'].endswith('.hidden') for p in data)

    head, *data, footer = run_rmlint('--partial-hidden -D -S a')
    data = filter_part_of_directory(data)
    assert len(data) == 2
    assert data[0]['path'].endswith('a')
    assert data[0]['type'] == 'duplicate_dir'
    assert data[1]['path'].endswith('b')

    head, *data, footer = run_rmlint('-D --no-partial-hidden -S a')
    data = filter_part_of_directory(data)
    assert len(data) == 0
